
__all__ = ['variables', 'calibrations', 'DataSet', 'PipelineCounter']

import os
import sys
import glob
import time
import numpy as np
import shutil
import datetime
import pylightcurve as plc
import astropy.io.fits as pf

from iraclis.__errors__ import *
from iraclis.__databases__ import iraclis_data


class Variable:

    def __init__(self, name, keyword=None, value=None, instance=None, kind='p', hdu_image=False, auto_fill=False):

        self.name = name
        if keyword is None:
            keyword = name
        self.keyword = keyword
        self.default_value = value
        self.value = value
        self.instance = instance
        self.kind = kind
        self.hdu_image = hdu_image
        self.auto_fill = auto_fill

    def set(self, value):

        if value is not None:

            if self.instance is None:
                self.value = value

            elif self.hdu_image:
                self.value = value

            elif self.auto_fill:
                if isinstance(value, str):
                    if value in ['auto', 'default', 'default_high', 'default_low', 'default_vlow']:
                        self.value = value
                    else:
                        try:
                            self.value = self.instance(value)
                        except ValueError:
                            print(value)
                            raise IraclisInputError('Input {0} parameter is not valid, '
                                                    '{1} is expected.'.format(self.name, self.instance))
                else:
                    try:
                        self.value = self.instance(value)
                    except ValueError:
                        raise IraclisInputError('Input {0} parameter is not valid, '
                                                '{1} is expected.'.format(self.name, self.instance))
            elif self.instance is bool:

                if isinstance(value, bool):
                    self.value = value
                elif isinstance(value, str):
                    if value == 'True':
                        self.value = True
                    elif value == 'False':
                        self.value = False
                else:
                    raise IraclisInputError('Input {0} parameter is not valid, '
                                            '{1} is expected.'.format(self.name, self.instance))
            else:
                try:
                    self.value = self.instance(value)
                except ValueError:
                        raise IraclisInputError('Input {0} parameter is not valid, '
                                                '{1} is expected.'.format(self.name, self.instance))

    def reset(self):

        self.set(self.default_value)

    def from_fits(self, fits, position=0):

        if self.hdu_image:
            self.set(fits[self.keyword].data)
        else:
            self.set(fits[position].header[self.keyword])

    def to_fits(self, fits, position=0, value=None):

        if value is not None:
            self.set(value)

        if self.hdu_image:
            try:
                fits[self.keyword] = pf.ImageHDU(name=self.keyword, header=fits[self.keyword].header, data=self.value)
            except KeyError:
                fits.append(pf.ImageHDU(name=self.keyword, data=self.value))
        else:
            fits[position].header.set(self.keyword, self.value)

    def from_dictionary(self, dictionary, sub_dictionary=None):

        if isinstance(dictionary, Variable):
            if sub_dictionary is None:
                self.set(dictionary.value[self.keyword])
            else:
                self.set(dictionary.value[sub_dictionary][self.keyword])
        else:
            if sub_dictionary is None:
                self.set(dictionary[self.keyword])
            else:
                self.set(dictionary[sub_dictionary][self.keyword])

    def to_dictionary(self, dictionary, sub_dictionary=None, value=None):

        if value is not None:
            self.set(value)

        if isinstance(dictionary, Variable):
            if sub_dictionary is None:
                dictionary.value[self.keyword] = self.value
            else:
                dictionary.value[sub_dictionary][self.keyword] = self.value
        else:
            if sub_dictionary is None:
                dictionary[self.keyword] = self.value
            else:
                dictionary[sub_dictionary][self.keyword] = self.value

    def custom(self, value=None):

        if value is None:
            return Variable(self.name, self.keyword, self.value, self.instance, self.kind,
                            self.hdu_image, self.auto_fill)
        else:
            return Variable(self.name, self.keyword, value, self.instance, self.kind, self.hdu_image,
                            self.auto_fill)

    def custom_from_fits(self, fits, position=0):

        x = self.custom()
        x.from_fits(fits, position)

        return x

    def custom_from_dictionary(self, dictionary, sub_dictionary=None):

        x = self.custom()
        x.from_dictionary(dictionary, sub_dictionary)

        return x


class Variables:

    def __init__(self):

        # process visit
        self.data_directory = Variable(
            'data_directory', value='.', instance=str, kind='u')
        self.reduction = Variable(
            'reduction', value=True, instance=bool, kind='u')
        self.splitting = Variable(
            'splitting', value=True, instance=bool, kind='u')
        self.extraction = Variable(
            'extraction', value=False, instance=bool, kind='u')
        self.splitting_extraction = Variable(
            'splitting_extraction', value=True, instance=bool, kind='u')
        self.fitting_white = Variable(
            'fitting_white', value=True, instance=bool, kind='u')
        self.fitting_spectrum = Variable(
            'fitting_spectrum', value=True, instance=bool, kind='u')
        self.raw_data_directory = Variable(
            'raw_data_directory', value='raw_data', instance=str)
        self.reduced_data_directory = Variable(
            'reduced_data_directory', value='reduced_data', instance=str)
        self.splitted_data_directory = Variable(
            'splitted_data_directory', value='splitted_data', instance=str)
        self.figures_directory = Variable(
            'figures_directory', value='figures', instance=str)
        self.output_directory = Variable(
            'output_directory', value='results', instance=str)
        self.output_directory_copy = Variable(
            'output_directory_copy', value='False', instance=str, kind='u')
        self.light_curve_file = Variable(
            'light_curve_file', value='extracted_light_curves', instance=str)
        self.fitting_file = Variable(
            'fitting_file', value='fitting_results', instance=str)
        self.apply_bias = Variable(
            'apply_bias', value=True, instance=bool)
        self.apply_linearity = Variable(
            'apply_linearity', value=True, instance=bool)
        self.apply_dark = Variable(
            'apply_dark', value=True, instance=bool)
        self.apply_gain = Variable(
            'apply_gain', value=True, instance=bool)
        self.apply_sky = Variable(
            'apply_sky', value=True, instance=bool)
        self.apply_flat = Variable(
            'apply_flat', value=True, instance=bool)
        self.apply_bpcr = Variable(
            'apply_bpcr', value=True, instance=bool)

        # observation
        self.ra_target = Variable(
            'ra_target', keyword='RA_TARG', instance=float)
        self.dec_target = Variable(
            'dec_target', keyword='DEC_TARG', instance=float)
        self.observation_type = Variable(
            'observation_type', keyword='OBSTYPE', instance=str)
        self.grism = Variable(
            'grism', keyword='FILTER', instance=str)
        self.wfc3_aperture = Variable(
            'wfc3_aperture', keyword='APERTURE', instance=str)
        self.postarg1 = Variable(
            'postarg1', keyword='POSTARG1', instance=float)
        self.scan_direction = Variable(
            'scan_direction', keyword='POSTARG2', instance=float)
        self.exposure_start = Variable(
            'exposure_start', keyword='EXPSTART', instance=float)
        self.exposure_end = Variable(
            'exposure_end', keyword='EXPEND', instance=float)
        self.exposure_time = Variable(
            'exposure_time', keyword='EXPTIME', instance=float)
        self.sub_array_type = Variable(
            'sub_array_type', keyword='SUBTYPE', instance=str)
        self.sub_array_size = Variable(
            'sub_array_size', keyword='SUBARRAY', instance=int)
        self.total_samples = Variable(
            'total_samples', keyword='NSAMP', instance=int)
        self.sampling_sequence = Variable(
            'sampling_sequence', keyword='SAMP_SEQ', instance=str)
        self.sample_number = Variable(
            'sample_number', keyword='SAMPNUM', instance=int)
        self.sample_time = Variable(
            'sample_time', keyword='SAMPTIME', instance=float)

        # timing
        self.heliocentric_julian_date = Variable(
            'heliocentric_julian_date', keyword='HJD', instance=float)

        # bias and zero-read
        self.zero_read = Variable(
            'zero_read', keyword='ZREAD', instance=np.array, hdu_image=True)
        self.zero_read_error = Variable(
            'zero_read_error', keyword='ZREADE', instance=np.array, hdu_image=True)
        self.zero_read_flux = Variable(
            'zero_read_flux', keyword='ZFLUX', instance=np.array, hdu_image=True)
        self.zero_read_flux_error = Variable(
            'zero_read_flux_error', keyword='ZFLUXE', instance=np.array, hdu_image=True)
        self.reference_pixels_level = Variable(
            'reference_pixels_level', keyword='REFLEV', instance=float)

        # sky
        self.sky_detection_limit = Variable(
            'sky_detection_limit', keyword='SKYLIM', value=2.0, instance=float)
        self.sky_background_level = Variable(
            'sky_background_level', keyword='SKYBGLEV', instance=float)
        self.sky_area = Variable(
            'sky_area', keyword='SKYAREA', instance=np.array, hdu_image=True)

        # bpcr
        self.cr_neighbours = Variable(
            'cr_neighbours', keyword='CRXNB', value=6, instance=int)
        self.cr_detection_limit = Variable(
            'cr_detection_limit', keyword='CRLIM', value=3.0, instance=float)
        self.use_bpcr_fast_mode = Variable(
            'use_bpcr_fast_mode', keyword='FBPCR', value=True, instance=bool)
        self.bpcr_map = Variable(
            'bpcr_map', keyword='BPCRMAP', instance=np.array, hdu_image=True)

        # calibration

        self.reference_pixel_x = Variable('reference_pixel_x', keyword='CRPIX1', instance=int)
        self.reference_pixel_y = Variable('reference_pixel_y', keyword='CRPIX2', instance=int)

        self.comparison_index_forward = Variable(
            'comparison_index_forward', keyword='FCOMP', value=0, instance=int)
        self.comparison_index_reverse = Variable(
            'comparison_index_reverse', keyword='RCOMP', value=1, instance=int)
        self.target_x_offset = Variable(
            'target_x_offset', keyword='TARXOFF', value=0, instance=float, kind='u')
        self.target_y_offset = Variable(
            'target_y_offset', keyword='TARYOFF', value=0, instance=float, kind='u')
        self.use_standard_flat = Variable(
            'use_standard_flat', keyword='STDFLAT', value=True, instance=bool)

        self.spectrum_bottom = Variable(
            'spectrum_bottom', keyword='SPCBTTM', instance=int)
        self.spectrum_top = Variable(
            'spectrum_top', keyword='SPCTOP', instance=int)
        self.spectrum_left = Variable(
            'spectrum_left', keyword='SPCLEFT', instance=int)
        self.spectrum_right = Variable(
            'spectrum_right', keyword='SPCRIGHT', instance=int)
        self.spectrum_scan = Variable(
            'spectrum_scan', keyword='SPCSCAN', instance=bool)
        self.spectrum_direction = Variable(
            'spectrum_direction', keyword='SPECDIR', instance=int)
        self.first_spectrum_bottom = Variable(
            'first_spectrum_bottom', keyword='FSPCBTTM', instance=int)
        self.first_spectrum_top = Variable(
            'first_spectrum_top', keyword='FSPCTOP', instance=int)
        self.first_spectrum_scan = Variable(
            'first_spectrum_scan', keyword='FSPCSCAN', instance=bool)
        self.first_spectrum_direction = Variable(
            'first_spectrum_direction', keyword='FSPECDIR', instance=int)

        self.comparison_x_star = Variable(
            'comparison_x_star', keyword='CMPXSTAR', instance=float)
        self.x_star = Variable(
            'x_star', keyword='XSTAR', instance=float)
        self.x_shift = Variable(
            'x_shift', keyword='XSHIFT', instance=float)
        self.x_shift_error = Variable(
            'x_shift_error', keyword='XSHIFTE', instance=float)

        self.comparison_y_star = Variable(
            'comparison_y_star', keyword='CMPYSTAR', instance=float)
        self.y_star = Variable(
            'y_star', keyword='YSTAR', instance=float)
        self.y_shift = Variable(
            'y_shift', keyword='YSHIFT', instance=float)
        self.y_shift_error = Variable(
            'y_shift_error', keyword='YSHIFTE', instance=float)

        self.scan_length = Variable(
            'scan_length', keyword='LEN', instance=float)
        self.scan_length_error = Variable(
            'scan_length_error', keyword='LEN_ERR', instance=float)

        self.wdpt_constant_coefficient_1 = Variable(
            'wdpt_constant_coefficient_1', keyword='CSCOEFF1', instance=float)
        self.wdpt_constant_coefficient_2 = Variable(
            'wdpt_constant_coefficient_2', keyword='CSCOEFF2', instance=float)
        self.wdpt_constant_coefficient_3 = Variable(
            'wdpt_constant_coefficient_3', keyword='CSCOEFF3', instance=float)
        self.wdpt_slope_coefficient_1 = Variable(
            'wdpt_slope_coefficient_1', keyword='SLCOEFF1', instance=float)
        self.wdpt_slope_coefficient_2 = Variable(
            'wdpt_slope_coefficient_2', keyword='SLCOEFF2', instance=float)
        self.wdpt_slope_coefficient_3 = Variable(
            'wdpt_slope_coefficient_3', keyword='SLCOEFF3', instance=float)

        self.scan_frame = Variable(
            'scan_frame', keyword='SCANMAP', instance=np.array, hdu_image=True)
        self.wavelength_frame = Variable(
            'wavelength_frame', keyword='WMAP', instance=np.array, hdu_image=True)
        self.normalised_wavelength_frame = Variable(
            'normalised_wavelength_frame', keyword='NWMAP', instance=np.array, hdu_image=True)

        # extraction
        self.aperture_lower_extend = Variable(
            'aperture_lower_extend', value=-20.0, instance=float, kind='u')
        self.aperture_upper_extend = Variable(
            'aperture_upper_extend', value=20.0, instance=float, kind='u')
        self.extraction_method = Variable(
            'extraction_method', value='gauss', instance=str, kind='u')
        self.extraction_gauss_sigma = Variable(
            'extraction_gauss_sigma', value=47.0, instance=float, kind='u')
        self.white_lower_wavelength = Variable(
            'white_lower_wavelength', value='default', instance=float, kind='u', auto_fill=True)
        self.white_upper_wavelength = Variable(
            'white_upper_wavelength', value='default', instance=float, kind='u', auto_fill=True)
        self.bins_file = Variable(
            'bins_file', value='default_high', instance=str, kind='u')
        self.bins_number = Variable(
            'bins_number', value=0, instance=int)
        self.heliocentric_julian_date_array = Variable(
            'heliocentric_julian_date_array', value=np.array([]), instance=np.array)
        self.spectrum_direction_array = Variable(
            'spectrum_direction_array', value=np.array([]), instance=np.array)
        self.sky_background_level_array = Variable(
            'sky_background_level_array', value=np.array([]), instance=np.array)
        self.x_star_array = Variable(
            'x_star_array', value=np.array([]), instance=np.array)
        self.x_shift_error_array = Variable(
            'x_shift_error_array', value=np.array([]), instance=np.array)
        self.y_star_array = Variable(
            'y_star_array', value=np.array([]), instance=np.array)
        self.y_shift_error_array = Variable(
            'y_shift_error_array', value=np.array([]), instance=np.array)
        self.scan_length_array = Variable(
            'scan_length_array', value=np.array([]), instance=np.array)
        self.scan_length_error_array = Variable(
            'scan_length_error_array', value=np.array([]), instance=np.array)
        self.lower_wavelength = Variable(
            'lower_wavelength', instance=float)
        self.upper_wavelength = Variable(
            'upper_wavelength', instance=float)
        self.ldc1 = Variable(
            'ldc1', instance=float, auto_fill=True)
        self.ldc2 = Variable(
            'ldc2', instance=float, auto_fill=True)
        self.ldc3 = Variable(
            'ldc3', instance=float, auto_fill=True)
        self.ldc4 = Variable(
            'ldc4', instance=float, auto_fill=True)
        self.flux_array = Variable(
            'flux', value=np.array([]), instance=np.array)
        self.error_array = Variable(
            'error', value=np.array([]), instance=np.array)
        self.ph_error_array = Variable(
            'ph_error', value=np.array([]), instance=np.array)
        self.white_dictionary = Variable(
            'white', value={}, instance=dict)
        self.light_curve_split = Variable(
            'light_curve_split', 'split_', instance=str)

        # fitting
        self.apply_up_down_stream_correction = Variable('apply_up_down_stream_correction', 'USDS', True, bool, 'u')
        self.exclude_initial_orbits = Variable('exclude_initial_orbits', 'EXCL1', 1, int, 'u')
        self.exclude_final_orbits = Variable('exclude_final_orbits', 'EXCL2', 0, int, 'u')
        self.exclude_initial_orbit_points = Variable('exclude_initial_orbit_points', 'EXCP1', 0, int, 'u')
        self.white_ldc1 = Variable('white_ldc1', 'WHTLDC1', 'default', float, 'u', auto_fill=True)
        self.white_ldc2 = Variable('white_ldc2', 'WHTLDC2', 'default', float, 'u', auto_fill=True)
        self.white_ldc3 = Variable('white_ldc3', 'WHTLDC3', 'default', float, 'u', auto_fill=True)
        self.white_ldc4 = Variable('white_ldc4', 'WHTLDC4', 'default', float, 'u', auto_fill=True)
        self.fit_ldc1 = Variable('fit_ldc1', 'FITLDC1', False, bool, 'u')
        self.fit_ldc2 = Variable('fit_ldc2', 'FITLDC2', False, bool, 'u')
        self.fit_ldc3 = Variable('fit_ldc3', 'FITLDC3', False, bool, 'u')
        self.fit_ldc4 = Variable('fit_ldc4', 'FITLDC4', False, bool, 'u')
        self.planet = Variable('planet', 'PLANET', 'auto', str, 'u', auto_fill=True)
        self.method = Variable('method', 'METHOD', 'claret', str, 'u')
        self.star_teff = Variable('star_teff', 'TEFF', 'auto', float, 'u', auto_fill=True)
        self.star_logg = Variable('star_logg', 'LOGG', 'auto', float, 'u', auto_fill=True)
        self.star_meta = Variable('star_meta', 'META', 'auto', float, 'u', auto_fill=True)
        self.rp_over_rs = Variable('rp_over_rs ', 'RP', 'auto', float, 'u', auto_fill=True)
        self.fp_over_fs = Variable('fp_over_fs', 'FP', 'auto', float, 'u', auto_fill=True)
        self.period = Variable('period', 'P', 'auto', float, 'u', auto_fill=True)
        self.sma_over_rs = Variable('sma_over_rs', 'A', 'auto', float, 'u', auto_fill=True)
        self.fit_sma_over_rs = Variable('fit_sma_over_rs', 'FITA', False, bool, 'u')
        self.eccentricity = Variable('eccentricity', 'E', 'auto', float, 'u', auto_fill=True)
        self.inclination = Variable('inclination', 'I', 'auto', float, 'u', auto_fill=True)
        self.fit_inclination = Variable('fit_inclination', 'FITI', False, bool, 'u')
        self.periastron = Variable('periastron', 'W', 'auto', float, 'u', auto_fill=True)
        self.mid_time = Variable('mid_time', 'MT', 'auto', float, 'u', auto_fill=True)
        self.fit_mid_time = Variable('fit_mid_time', 'FITMT', 'auto', bool, 'u', auto_fill=True)
        self.second_order_ramp = Variable('second_order_ramp', 'RAMP2', False, bool, 'u')
        self.first_orbit_ramp = Variable('first_orbit_ramp', 'FOR', True, bool, 'u')
        self.mid_orbit_ramps = Variable('mid_orbit_ramps', 'MOR', True, bool, 'u')
        self.mcmc_iterations = Variable('mcmc_iterations', 'MCMCI', 300000, int, 'u')
        self.mcmc_walkers = Variable('mcmc_walkers', 'MCMCW', 200, int, 'u')
        self.mcmc_burned_iterations = Variable('mcmc_burned_iterations', 'MCMCB', 200000, int, 'u')
        self.spectral_mcmc_iterations = Variable('spectral_mcmc_iterations', 'SPMCMCI', 50000, int, 'u')
        self.spectral_mcmc_walkers = Variable('spectral_mcmc_walkers', 'SPMCMCW', 100, int, 'u')
        self.spectral_mcmc_burned_iterations = Variable('spectral_mcmc_burned_iterations', 'SPMCMCB', 20000, int, 'u')

    def from_parameters_file(self, parameters_file=None, data_directory=None):

        # check user inputs

        variables.reset()

        if parameters_file and data_directory:

            raise IraclisInputError('You should give as input either a parameters file '
                                    '(which includes a data directory path), '
                                    'OR a data directory path '
                                    '(if you want to run in default parameters).')

        elif parameters_file:

            parameters_file = os.path.abspath(parameters_file)
            if not os.path.isfile(parameters_file):
                raise IraclisFileError('No such file: ' + parameters_file)
            print('Loading parameters file: {} ...'.format(os.path.split(parameters_file)[1]))

            parameters_in_file = []

            for i in open(parameters_file).readlines():
                if len(i.split()) > 0:
                    if i.split()[0][0] != '#':
                        if i.split()[0] in vars(self):
                            parameters_in_file.append(i.split()[0])
                            vars(self)[i.split()[0]].set(i.split()[1])

            for i in vars(self):
                if not isinstance(vars(self)[i], list):
                    if i not in parameters_in_file:
                        if vars(self)[i].kind == 'u':
                            print('WARNING: Parameter not in file, setting {0}={1}'.format(
                                i, vars(self)[i].default_value))

        elif data_directory:

            self.data_directory.set(data_directory)

        else:

            raise IraclisInputError('You should give as input either a parameters file '
                                    '(which includes a data directory path), '
                                    'OR a data directory path '
                                    '(if you want to run in default parameters).')

    def overwrite(self, procedure=None, par_string=None):

        if procedure:
            self.reduction.set(bool(int(procedure[0])))
            self.splitting.set(bool(int(procedure[1])))
            self.extraction.set(bool(int(procedure[2])))
            self.splitting_extraction.set(bool(int(procedure[3])))
            self.fitting_white.set(bool(int(procedure[4])))
            self.fitting_spectrum.set(bool(int(procedure[5])))

        if par_string:
            for i in par_string.split(','):
                if len(i.split('=')) > 0:
                    if i.split('=')[0] in vars(self):
                        vars(self)[i.split('=')[0]].set(i.split('=')[1])

    def set_binning(self, input_data, white_lower_wavelength, white_upper_wavelength, white_ldc1,
                    white_ldc2, white_ldc3, white_ldc4, bins_file):

        grism = self.grism.custom()
        sub_array_size = self.sub_array_size.custom()

        if isinstance(input_data, dict):
            grism.from_dictionary(input_data)
            sub_array_size.from_dictionary(input_data)
        elif input_data.splitted:
            grism.from_fits(input_data.spectroscopic_images[0][0])
            sub_array_size.from_fits(input_data.spectroscopic_images[0][0])
        else:
            grism.from_fits(input_data.spectroscopic_images[0])
            sub_array_size.from_fits(input_data.spectroscopic_images[0])

        if grism.value == 'G141':
            if sub_array_size.value < 200:
                if white_lower_wavelength == 'default':
                    white_lower_wavelength = 11000
                if white_upper_wavelength == 'default':
                    white_upper_wavelength = 16500
            else:
                if white_lower_wavelength == 'default':
                    white_lower_wavelength = 10800
                if white_upper_wavelength == 'default':
                    white_upper_wavelength = 16800

        elif grism.value == 'G102':
            if white_lower_wavelength == 'default':
                white_lower_wavelength = 8000
            if white_upper_wavelength == 'default':
                white_upper_wavelength = 11500

        if os.path.isfile(os.path.abspath(bins_file)):
            bins_file = os.path.abspath(bins_file)
            print('Loading bins file: {} ...'.format(os.path.split(bins_file)[1]))

            if len(np.loadtxt(bins_file, unpack=True)) == 6:
                bins_lower_wavelength, bins_upper_wavelength, bins_ldc1, bins_ldc2, bins_ldc3, bins_ldc4 = \
                    np.loadtxt(bins_file, unpack=True)
                bins_ldc1 = list(bins_ldc1)
                bins_ldc2 = list(bins_ldc2)
                bins_ldc3 = list(bins_ldc3)
                bins_ldc4 = list(bins_ldc4)

            else:
                raise IraclisFileError('Not valid bins file.')

        elif bins_file == 'default_high':
            print('Loading bins file: {} ...'.format('default_high.txt'))
            if grism.value == 'G141':
                bins_file = os.path.join(iraclis_data.hstwfc3(), 'default_high.txt')
            elif grism.value == 'G102':
                bins_file = os.path.join(iraclis_data.hstwfc3(), 'default_high_g102.txt')

            bins_lower_wavelength, bins_upper_wavelength = np.loadtxt(bins_file, unpack=True)
            bins_ldc1 = ['default_high' for ff in bins_lower_wavelength]
            bins_ldc2 = ['default_high' for ff in bins_lower_wavelength]
            bins_ldc3 = ['default_high' for ff in bins_lower_wavelength]
            bins_ldc4 = ['default_high' for ff in bins_lower_wavelength]

        elif bins_file == 'default_low':
            print('Loading bins file: {} ...'.format('default_low.txt'))
            if grism.value == 'G141':
                bins_file = os.path.join(iraclis_data.hstwfc3(), 'default_low.txt')
            elif grism.value == 'G102':
                bins_file = os.path.join(iraclis_data.hstwfc3(), 'default_low_g102.txt')

            bins_lower_wavelength, bins_upper_wavelength = np.loadtxt(bins_file, unpack=True)
            bins_ldc1 = ['default_low' for ff in bins_lower_wavelength]
            bins_ldc2 = ['default_low' for ff in bins_lower_wavelength]
            bins_ldc3 = ['default_low' for ff in bins_lower_wavelength]
            bins_ldc4 = ['default_low' for ff in bins_lower_wavelength]

        elif bins_file == 'default_vlow':
            print('Loading bins file: {} ...'.format('default_vlow.txt'))
            if grism.value == 'G141':
                bins_file = os.path.join(iraclis_data.hstwfc3(), 'default_vlow.txt')
            elif grism.value == 'G102':
                bins_file = os.path.join(iraclis_data.hstwfc3(), 'default_vlow_g102.txt')

            bins_lower_wavelength, bins_upper_wavelength = np.loadtxt(bins_file, unpack=True)
            bins_ldc1 = ['default_vlow' for ff in bins_lower_wavelength]
            bins_ldc2 = ['default_vlow' for ff in bins_lower_wavelength]
            bins_ldc3 = ['default_vlow' for ff in bins_lower_wavelength]
            bins_ldc4 = ['default_vlow' for ff in bins_lower_wavelength]

        else:
            raise IraclisFileError('No bins file found')

        lower_wavelength = self.lower_wavelength.custom()
        upper_wavelength = self.upper_wavelength.custom()
        flux_array = self.flux_array.custom()
        error_array = self.error_array.custom()
        ph_error_array = self.ph_error_array.custom()
        ldc1 = self.ldc1.custom()
        ldc2 = self.ldc2.custom()
        ldc3 = self.ldc3.custom()
        ldc4 = self.ldc4.custom()

        white_dictionary = Variable('white', value={}, instance=dict)
        bins_dictionaries = [Variable('bin_{0}'.format(str(ff + 1).zfill(2)), value={}, instance=dict)
                             for ff in range(len(bins_lower_wavelength))]

        for i, j in enumerate([white_dictionary] + bins_dictionaries):
            flux_array.to_dictionary(j)
            error_array.to_dictionary(j)
            ph_error_array.to_dictionary(j)
            lower_wavelength.to_dictionary(j, value=([white_lower_wavelength] + list(bins_lower_wavelength))[i])
            upper_wavelength.to_dictionary(j, value=([white_upper_wavelength] + list(bins_upper_wavelength))[i])
            ldc1.to_dictionary(j, value=([white_ldc1] + bins_ldc1)[i])
            ldc2.to_dictionary(j, value=([white_ldc2] + bins_ldc2)[i])
            ldc3.to_dictionary(j, value=([white_ldc3] + bins_ldc3)[i])
            ldc4.to_dictionary(j, value=([white_ldc4] + bins_ldc4)[i])

        return white_dictionary, bins_dictionaries

    def reset(self):

        for i in vars(self):
            if isinstance(vars(self)[i], list):
                for j in vars(self)[i]:
                    j.reset()
            else:
                vars(self)[i].reset()

    def save(self, export_file):

        w = open(export_file, 'w')

        for i in vars(self):
            if not isinstance(vars(self)[i], list):
                if vars(self)[i].kind == 'u':
                    w.write('{0}{1}{2}\n'.format(i, ' ' * (35 - len(i)), vars(self)[i].value))

        w.close()


variables = Variables()


# calibration variables


class Calibration:

    def __init__(self, calibration_files_id, keyword, value=None, matching_keywords=None,
                 hdu_image=False, dark_mode=False):

        self.calibration_files_id = calibration_files_id
        self.calibration_data = None
        self.keyword = keyword
        self.value = value
        self.hdu_image = hdu_image
        self.dark_mode = dark_mode
        self.matching_keywords = matching_keywords

    def match(self, input_data):

        if isinstance(input_data, pf.HDUList):
            pass
        elif input_data.splitted:
            input_data = input_data.spectroscopic_images[0][0]
        else:
            input_data = input_data.spectroscopic_images[0]

        calibration_directory_path = iraclis_data.hstwfc3()

        if not self.calibration_data:

            self.calibration_data = \
                [pf.open(pp) for pp in
                 glob.glob(os.path.join(calibration_directory_path, '*{0}*.fits'.format(self.calibration_files_id)))]

        if len(self.calibration_data) == 1 or not self.matching_keywords:

            test = True
            selected_calibration_data = self.calibration_data[0]

        else:

            test = False
            selected_calibration_data = self.calibration_data[0]
            for selected_calibration_data in self.calibration_data:
                if ([input_data[0].header[pp] for pp in self.matching_keywords] ==
                        [selected_calibration_data[0].header[pp] for pp in self.matching_keywords]):
                    test = True
                    break

        if test:

            if self.hdu_image:

                if self.dark_mode:

                    dark_list = np.array([selected_calibration_data[0]])
                    sample_numbers = [input_data[pp].header[variables.sample_number.keyword]
                                      for pp in plc.fits_sci(input_data)]

                    for i in plc.fits_sci(selected_calibration_data):
                        if (selected_calibration_data[i].header[
                                variables.sample_number.keyword] in sample_numbers
                            or selected_calibration_data[i].header[
                                variables.sample_number.keyword] == min(sample_numbers) - 1):

                            dark_list = np.append(dark_list, selected_calibration_data[i])
                            dark_list = np.append(dark_list, selected_calibration_data[i + 1])
                            dark_list = np.append(dark_list, selected_calibration_data[i + 2])
                            dark_list = np.append(dark_list, selected_calibration_data[i + 3])
                            dark_list = np.append(dark_list, selected_calibration_data[i + 4])

                    return pf.HDUList(list(dark_list))

                else:

                    array = (np.ones_like(selected_calibration_data[self.keyword].data) *
                             selected_calibration_data[self.keyword].data)

                    crop1 = int(len(np.array(array)) / 2 - len(input_data[plc.fits_sci(input_data)[0]].data) / 2)
                    crop2 = int(len(np.array(array)) / 2 + len(input_data[plc.fits_sci(input_data)[0]].data) / 2)

                    return array[crop1:crop2, crop1:crop2]

            else:
                return (np.ones_like(selected_calibration_data[0].header[self.keyword]) *
                        selected_calibration_data[0].header[self.keyword])

        else:
            raise IraclisLibraryError('No suitable dark current calibration file found, \n'
                                      'you must update your library manually. Search at: \n\n'
                                      '\thttps://hst-crds.stsci.edu/ \n\n'
                                      'for the latest dark current calibration file for\n\n'
                                      '\tSAMP_SEQ: {0} \n\tSUBTYTE: {1}'.format(input_data[0].header['SAMP_SEQ'],
                                                                                input_data[0].header['SUBTYPE']))


class Calibrations:

    def __init__(self):

        self.super_zero_read = Calibration(
            'lin_custom', 'ZSCI', hdu_image=True)
        self.super_zero_read_error = Calibration(
            'lin_custom', 'ZERR', hdu_image=True)
        self.ccd_gain = Calibration(
            'ccd_custom', 'CCDGAIN', hdu_image=True)
        self.ccd_read_noise = Calibration(
            'ccd_custom', 'CCDREADN', hdu_image=True)
        self.linearity_coefficient_1 = Calibration(
            'lin_custom', 'LINC1', hdu_image=True)
        self.linearity_coefficient_2 = Calibration(
            'lin_custom', 'LINC2', hdu_image=True)
        self.linearity_coefficient_3 = Calibration(
            'lin_custom', 'LINC3', hdu_image=True)
        self.linearity_coefficient_4 = Calibration(
            'lin_custom', 'LINC4', hdu_image=True)
        self.super_dark = Calibration(
            'drk', 'SCI', dark_mode=True, hdu_image=True,
            matching_keywords=[variables.sampling_sequence.keyword, variables.sub_array_type.keyword])
        self.mean_gain = Calibration(
            'ccd_custom', 'MGAIN')
        self.gain_profile = Calibration(
            'pfl', 'SCI', hdu_image=True)
        self.master_sky = Calibration(
            'sky', 'PRIMARY', hdu_image=True, matching_keywords=[variables.grism.keyword])
        self.flat_field_coefficient_1 = Calibration(
            'flat', 'FFC1', hdu_image=True, matching_keywords=[variables.grism.keyword])
        self.flat_field_coefficient_2 = Calibration(
            'flat', 'FFC2', hdu_image=True, matching_keywords=[variables.grism.keyword])
        self.flat_field_coefficient_3 = Calibration(
            'flat', 'FFC3', hdu_image=True, matching_keywords=[variables.grism.keyword])
        self.flat_field_coefficient_4 = Calibration(
            'flat', 'FFC4', hdu_image=True, matching_keywords=[variables.grism.keyword])
        self.flat_field_min_wavelength = Calibration(
            'flat', 'WMIN', matching_keywords=[variables.grism.keyword])
        self.flat_field_max_wavelength = Calibration(
            'flat', 'WMAX', matching_keywords=[variables.grism.keyword])
        self.bad_pixels = Calibration(
            'bpx_custom', 'PRIMARY', hdu_image=True)
        self.trace_at0 = Calibration(
            'calib', 'AT0', matching_keywords=[variables.grism.keyword])
        self.trace_at1 = Calibration(
            'calib', 'AT1', matching_keywords=[variables.grism.keyword])
        self.trace_at2 = Calibration(
            'calib', 'AT2', matching_keywords=[variables.grism.keyword])
        self.trace_at3 = Calibration(
            'calib', 'AT3', matching_keywords=[variables.grism.keyword])
        self.trace_at4 = Calibration(
            'calib', 'AT4', matching_keywords=[variables.grism.keyword])
        self.trace_at5 = Calibration(
            'calib', 'AT5', matching_keywords=[variables.grism.keyword])
        self.trace_bt0 = Calibration(
            'calib', 'BT0', matching_keywords=[variables.grism.keyword])
        self.trace_bt1 = Calibration(
            'calib', 'BT1', matching_keywords=[variables.grism.keyword])
        self.trace_bt2 = Calibration(
            'calib', 'BT2', matching_keywords=[variables.grism.keyword])
        self.wsol_aw0 = Calibration(
            'calib', 'AW0', matching_keywords=[variables.grism.keyword])
        self.wsol_aw1 = Calibration(
            'calib', 'AW1', matching_keywords=[variables.grism.keyword])
        self.wsol_aw2 = Calibration(
            'calib', 'AW2', matching_keywords=[variables.grism.keyword])
        self.wsol_aw3 = Calibration(
            'calib', 'AW3', matching_keywords=[variables.grism.keyword])
        self.wsol_aw4 = Calibration(
            'calib', 'AW4', matching_keywords=[variables.grism.keyword])
        self.wsol_aw5 = Calibration(
            'calib', 'AW5', matching_keywords=[variables.grism.keyword])
        self.wsol_bw0 = Calibration(
            'calib', 'BW0', matching_keywords=[variables.grism.keyword])
        self.wsol_bw1 = Calibration(
            'calib', 'BW1', matching_keywords=[variables.grism.keyword])
        self.wsol_bw2 = Calibration(
            'calib', 'BW2', matching_keywords=[variables.grism.keyword])


calibrations = Calibrations()


class DataSet:

    def __init__(self, input_data, direct_image=None):

        if direct_image is not None:
            if isinstance(direct_image, str):
                if os.path.isfile(direct_image):
                    direct_image = pf.open(direct_image)
                else:
                    raise IraclisFileError('No such file {0}'.format(input_data))
            else:
                raise IraclisFileError('Please give a file name for the direct image or leave it empty.')

        if not isinstance(input_data, str):
            raise IraclisFileError('Please give a file or directory name for the input data.')

        elif input_data == 'empty':
            self.file_names = []
            self.spectroscopic_images = []
            self.direct_image = [1]
            self.splitted = False
            self._data_set_directory_path = None

        elif os.path.isfile(input_data):

            self.file_names = []
            self.spectroscopic_images = [pf.open(input_data)]
            if not direct_image:
                self.direct_image = None
            else:
                self.direct_image = direct_image
            self.splitted = False
            self._data_set_directory_path = None

            if self.spectroscopic_images[0][0].header[variables.observation_type.keyword] != 'SPECTROSCOPIC':
                raise IraclisFileError('A single direct image is not s valid input dataset.')

        elif os.path.isdir(input_data) and len(glob.glob(os.path.join(input_data, '*', ''))) == 0:

            nsamp = []
            final_list = []
            direct_image = None

            files = sorted(glob.glob(os.path.join(input_data, '*.fits')))
            for i in files:
                with pf.open(i) as j:
                    if j[0].header[variables.observation_type.keyword] == 'SPECTROSCOPIC':
                        final_list.append([j[0].header[variables.exposure_start.keyword],
                                           os.path.split(i)[1], plc.copy_fits(j)])
                        nsamp.append(j[0].header[variables.total_samples.keyword])
                    elif not direct_image:
                        direct_image = plc.copy_fits(j[:2])

            nsamps = [int(np.median(np.array(nsamp)))]

            final_list.sort()
            list_of_times, list_of_files, list_of_fits = np.swapaxes(np.array(final_list, dtype=object), 0, 1)

            outliers = True
            while outliers:
                outliers = False
                for i in range(len(list_of_fits)):
                    if list_of_fits[i][0].header[variables.total_samples.keyword] not in nsamps:
                        list_of_fits = np.delete(list_of_fits, i)
                        list_of_files = np.delete(list_of_files, i)
                        outliers = True
                        break

            self.file_names = list_of_files
            self.spectroscopic_images = list_of_fits
            self.direct_image = direct_image
            self.splitted = False
            self._data_set_directory_path = input_data

        elif os.path.isdir(input_data) and len(glob.glob(os.path.join(input_data, '*', ''))) > 0:

            self.file_names = []
            self.spectroscopic_images = []
            self.direct_image = []
            self.splitted = True
            self._data_set_directory_path = input_data

            for input_data in sorted(glob.glob(os.path.join(input_data, '*', ''))):

                nsamp = []
                final_list = []
                direct_image = None

                files = sorted(glob.glob(os.path.join(input_data, '*.fits')))
                for i in files:
                    with pf.open(i) as j:
                        if j[0].header[variables.observation_type.keyword] == 'SPECTROSCOPIC':
                            final_list.append([j[0].header[variables.exposure_start.keyword],
                                               os.path.split(i)[1], plc.copy_fits(j)])
                            nsamp.append(j[0].header[variables.total_samples.keyword])
                        elif not direct_image:
                            direct_image = plc.copy_fits(j[:2])

                nsamps = [int(np.median(np.array(nsamp)))]

                final_list.sort()
                list_of_times, list_of_files, list_of_fits = np.swapaxes(final_list, 0, 1)

                outliers = True
                while outliers:
                    outliers = False
                    for i in range(len(list_of_fits)):
                        if list_of_fits[i][0].header[variables.total_samples.keyword] not in nsamps:
                            list_of_fits = np.delete(list_of_fits, i)
                            list_of_files = np.delete(list_of_files, i)
                            outliers = True
                            break

                self.file_names = list_of_files
                self.spectroscopic_images.append(list_of_fits)
                self.direct_image = direct_image

        else:
            raise IraclisFileError('No such file or directory: {0}'.format(input_data))

        if not self.direct_image:
            raise IraclisFileError('A direct image is necessary.')

    def save(self, export_directory, arrange=True, export_pipeline_variables_file='variables.txt'):

        if os.path.isdir(export_directory):
            backup = '{0}_{1}'.format(export_directory, time.strftime('%y-%m-%d_%H-%M-%S'))
            shutil.copytree(export_directory, backup)
            shutil.rmtree(export_directory)

        os.mkdir(export_directory)

        if arrange:
            for i in range(len(self.file_names)):
                date = str(self.spectroscopic_images[i][0].header['DATE-OBS'])
                obs_time = str(self.spectroscopic_images[i][0].header['TIME-OBS'])
                obs_time = '-'.join(obs_time.split(':'))
                if self.file_names[i].split('_')[0] != date or self.file_names[i].split('_')[1] != obs_time:
                    self.file_names[i] = '{0}_{1}_{2}'.format(date, obs_time, os.path.split(self.file_names[i])[1])

        for i in range(len(self.file_names)):

            copy_of_file = plc.copy_fits(self.spectroscopic_images[i])
            try:
                copy_of_file.writeto(os.path.join(export_directory, self.file_names[i]), output_verify='fix')
            except pf.VerifyError as e:
                hdu = int(str(e.args)[4:-4].split('\\n')[1].replace('HDU ', '').replace(':', ''))
                card = int(str(e.args)[4:-4].split('\\n')[2].replace('Card ', '').replace(':', ''))
                del copy_of_file[hdu].header[card]
                copy_of_file.writeto(os.path.join(export_directory, self.file_names[i]), output_verify='fix')

        copy_of_file = plc.copy_fits(self.direct_image)
        try:
            copy_of_file.writeto(os.path.join(export_directory, 'direct_image.fits'), output_verify='fix')
        except pf.VerifyError as e:
            hdu = int(str(e.args)[4:-4].split('\\n')[1].replace('HDU ', '').replace(':', ''))
            card = int(str(e.args)[4:-4].split('\\n')[2].replace('Card ', '').replace(':', ''))
            del copy_of_file[hdu].header[card]
            copy_of_file.writeto(os.path.join(export_directory, 'direct_image.fits'), output_verify='fix')

        if export_pipeline_variables_file:
            variables.save(os.path.join(export_directory, export_pipeline_variables_file))

    def copy_split(self, split_number):

        x = DataSet('empty')
        x.spectroscopic_images = self.spectroscopic_images[split_number]

        return x


# pipeline counter


class PipelineCounter:

    def __init__(self, task, total_iterations, show_every=1):

        self.task = '{0}{1}'.format(task, '.' * (15 - len(task)))
        self.current_iteration = 0
        self.total_iterations = int(total_iterations)
        self.start_time = time.time()
        self.show = 0
        self.show_every = int(show_every)

        if self.total_iterations == 1:
            self.show_every = 10

    def update(self):

        self.current_iteration += 1
        self.show += 1.0 / self.show_every

        out_of = '{0}{1} / {2}'.format(' ' * (len(str(self.total_iterations)) - len(str(self.current_iteration))),
                                       str(self.current_iteration), str(self.total_iterations))

        delta_time = time.time() - self.start_time

        time_left = str(datetime.timedelta(
            seconds=int((self.total_iterations - self.current_iteration) * delta_time / self.current_iteration)))

        total_time = str(datetime.timedelta(seconds=int(delta_time)))

        if int(self.show):
            sys.stdout.write('\r\033[K')
            sys.stdout.write('{0}: {1}   time left: {2}   total time: {3}'.format(
                self.task, out_of, time_left, total_time))
            sys.stdout.flush()
            self.show = 0

        if self.current_iteration == self.total_iterations and self.total_iterations > 1:
            print('')




"""
images3_photometry.py

Includes all the functions that perform photometry processes.
All the functions take as input either an HDUList object or a DataSet object, as defined in the basics.py file and
return the input object and a dictionary that contains the extracted light-curves. In all cases, the default values for
the input parameters are the values in the respective pipeline.variables object. Note that the parameters for the
supporting functions do not have default values, as their purpose is to be used only in this particular file.

Functions included:
photometry:         ...
split_photometry:   ...

Supporting functions included:
get_flux_integral:  ...
get_flux_gauss:     ...

"""

__all__ = ['photometry', 'plot_photometry', 'split_photometry']

import numpy as np
import scipy
import warnings
import pylightcurve as plc

from matplotlib import pyplot as plt

from iraclis.classes import *


def get_flux_integral(fits, lower_wavelength, upper_wavelength,
                      aperture_lower_extend, aperture_upper_extend, sigma, plot=False):

    x_star = variables.x_star.custom_from_fits(fits).value
    y_star = variables.y_star.custom_from_fits(fits).value
    spectrum_direction = variables.spectrum_direction.custom_from_fits(fits).value
    scan_length = variables.scan_length.custom_from_fits(fits).value
    wdpt_constant_coefficient_1 = variables.wdpt_constant_coefficient_1.custom_from_fits(fits).value
    wdpt_constant_coefficient_2 = variables.wdpt_constant_coefficient_2.custom_from_fits(fits).value
    wdpt_constant_coefficient_3 = variables.wdpt_constant_coefficient_3.custom_from_fits(fits).value
    wdpt_slope_coefficient_1 = variables.wdpt_slope_coefficient_1.custom_from_fits(fits).value
    wdpt_slope_coefficient_2 = variables.wdpt_slope_coefficient_2.custom_from_fits(fits).value
    wdpt_slope_coefficient_3 = variables.wdpt_slope_coefficient_3.custom_from_fits(fits).value

    trace_at0 = calibrations.trace_at0.match(fits)
    trace_at1 = calibrations.trace_at1.match(fits)
    trace_at2 = calibrations.trace_at2.match(fits)
    trace_at3 = calibrations.trace_at3.match(fits)
    trace_at4 = calibrations.trace_at4.match(fits)
    trace_at5 = calibrations.trace_at5.match(fits)
    trace_bt0 = calibrations.trace_bt0.match(fits)
    trace_bt1 = calibrations.trace_bt1.match(fits)
    trace_bt2 = calibrations.trace_bt2.match(fits)

    def get_trace(dy):

        xx0 = x_star
        yy0 = y_star + dy
        sub = 507 - len(fits[1].data) / 2

        bt = trace_bt0 + trace_bt1 * xx0 + trace_bt2 * yy0
        at = (trace_at0 + trace_at1 * xx0 + trace_at2 * yy0 + trace_at3 * xx0 * xx0 +
              trace_at4 * xx0 * yy0 + trace_at5 * yy0 * yy0)
        return at, bt + yy0 - at * xx0 - sub + at * sub

    if spectrum_direction > 0:
        y0 = aperture_lower_extend
        y1 = scan_length + aperture_upper_extend
    else:
        y0 = - scan_length - aperture_upper_extend
        y1 = - aperture_lower_extend

    va1 = (wdpt_slope_coefficient_1 / (wdpt_slope_coefficient_2 + lower_wavelength) + wdpt_slope_coefficient_3)
    vb1 = (wdpt_constant_coefficient_1 / (wdpt_constant_coefficient_2 + lower_wavelength) + wdpt_constant_coefficient_3)
    va2 = (wdpt_slope_coefficient_1 / (wdpt_slope_coefficient_2 + upper_wavelength) + wdpt_slope_coefficient_3)
    vb2 = (wdpt_constant_coefficient_1 / (wdpt_constant_coefficient_2 + upper_wavelength) + wdpt_constant_coefficient_3)
    ha1, hb1 = get_trace(y0)
    ha2, hb2 = get_trace(y1)

    ha2 += sigma
    ha2 -= sigma

    if plot:
        xxx = np.arange((hb1 - vb1) / (va1 - ha1), (hb1 - vb2) / (va2 - ha1), 0.0001)
        plt.plot(xxx, ha1 * xxx + hb1, 'w-')
        xxx = np.arange((hb2 - vb1) / (va1 - ha2), (hb2 - vb2) / (va2 - ha2), 0.0001)
        plt.plot(xxx, ha2 * xxx + hb2, 'w-')
        xxx = np.arange((hb2 - vb1) / (va1 - ha2), (hb1 - vb1) / (va1 - ha1), 0.0001)
        plt.plot(xxx, va1 * xxx + vb1, 'w-')
        xxx = np.arange((hb2 - vb2) / (va2 - ha2), (hb1 - vb2) / (va2 - ha1), 0.0001)
        plt.plot(xxx, va2 * xxx + vb2, 'w-')

    fcr = np.full_like(fits[1].data, fits[1].data)
    fhm = np.roll(fcr, 1, axis=1)
    fhp = np.roll(fcr, -1, axis=1)
    fvm = np.roll(fcr, -1, axis=0)
    fvp = np.roll(fcr, 1, axis=0)
    x0, y0 = np.meshgrid(np.arange(len(fcr)), np.arange(len(fcr)))

    summ1 = (2.0 * fcr - fhm - fhp)
    summ2 = (4.0 * fcr - 4.0 * fhm)
    summ3 = (8.0 * fcr + 4.0 * fhm - 2.0 * fhp + 4.0 * fvm - 2.0 * fvp)
    summ4 = (4.0 * fcr - 4.0 * fvm)
    summ5 = (2.0 * fcr - fvm - fvp)

    summ6 = (4.0 * fcr - 4.0 * fhp)
    summ7 = (10.0 * fcr - 2.0 * fhm + 4.0 * fhp - fvm + fvp)
    summ8 = (20.0 * fcr - 4.0 * fhm + 8.0 * fhp)

    summ9 = (8.0 * fcr - 2.0 * fhm + 4.0 * fhp - 2.0 * fvm + 4.0 * fvp)
    summ10 = (4.0 * fcr - 4.0 * fvp)
    summ11 = (2.0 * fcr - fvm - fvp)

    # left edge
    a, b = va1, vb1
    x1 = (-b + y0) / a
    x2 = (1 - b + y0) / a

    formula = a * x0 + b - y0
    formula_2 = formula * formula
    formula_3 = formula_2 * formula
    formula_4 = formula_3 * formula

    case1 = (
        + fcr - (1.0 / (24.0 * (a ** 3))) * (
            + summ1 * formula_4
            + a * summ2 * formula_3
            + (a ** 2) * (- summ3 * formula_2 - summ4 * formula_3 + summ5 * formula_4)
        )
    )

    formula = a + a * x0 + b - y0
    formula_2 = formula * formula
    formula_3 = formula_2 * formula

    case2 = (
        - (1.0 / (24.0 * (a ** 3))) * (
            + 4.0 * summ1 * (- 0.25 + formula - 1.5 * formula_2 + formula_3)
            + (a * 3.0) * summ6 * (-1.0 / 3 + formula - formula_2)
            + (a ** 2) * (summ7 - summ8 * formula)
        )
    )

    formula = - 1.0 + a + a * x0 + b - y0
    formula_2 = formula * formula
    formula_3 = formula_2 * formula
    formula_4 = formula_3 * formula

    case3 = (
        - (1.0 / (24.0 * (a ** 3))) * (
            - summ1 * formula_4
            + a * summ6 * formula_3
            + (a ** 2) * (summ9 * formula_2 - summ10 * formula_3 - summ11 * formula_4)
        )
    )

    new_data = np.full_like(fits[1].data, fits[1].data)
    new_data = np.where((x1 > x0) & (x2 < x0), case1, new_data)
    new_data = np.where((x1 > x0) & (x0 + 1 > x1) & (x2 > x0) & (x0 + 1 > x2), case2, new_data)
    new_data = np.where((x1 > x0 + 1) & (x2 < x0 + 1), case3, new_data)
    new_data = np.where((x1 > x0 + 1) & (x2 > x0 + 1), 0, new_data)

    # right edge
    a, b = va2, vb2
    x1 = (-b + y0) / a
    x2 = (1 - b + y0) / a

    formula = a * x0 + b - y0
    formula_2 = formula * formula
    formula_3 = formula_2 * formula
    formula_4 = formula_3 * formula

    case1 = (
        + (1.0 / (24.0 * (a ** 3))) * (
            + summ1 * formula_4
            + a * summ2 * formula_3
            + (a ** 2) * (- summ3 * formula_2 - summ4 * formula_3 + summ5 * formula_4)
        )
    )

    formula = a + a * x0 + b - y0
    formula_2 = formula * formula
    formula_3 = formula_2 * formula

    case2 = (
        fcr + (1.0 / (24.0 * (a ** 3))) * (
            + 4.0 * summ1 * (- 0.25 + formula - 1.5 * formula_2 + formula_3)
            + (a * 3.0) * summ6 * (-1.0 / 3 + formula - formula_2)
            + (a ** 2) * (summ7 - summ8 * formula)
        )
    )

    formula = - 1.0 + a + a * x0 + b - y0
    formula_2 = formula * formula
    formula_3 = formula_2 * formula
    formula_4 = formula_3 * formula

    case3 = (
        fcr + (1.0 / (24.0 * (a ** 3))) * (
            - summ1 * formula_4
            + a * summ6 * formula_3
            + (a ** 2) * (summ9 * formula_2 - summ10 * formula_3 - summ11 * formula_4)
        )
    )

    new_data = np.where((x1 < x0) & (x2 < x0), 0, new_data)
    new_data = np.where((x1 > x0) & (x2 < x0), case1, new_data)
    new_data = np.where((x1 > x0) & (x0 + 1 > x1) & (x2 > x0) & (x0 + 1 > x2), case2, new_data)
    new_data = np.where((x1 > x0 + 1) & (x2 < x0 + 1), case3, new_data)

    # upper edge

    new_data = np.rot90(new_data)

    fcr = np.ones_like(new_data) * new_data
    fhm = np.roll(fcr, 1, axis=1)
    fhp = np.roll(fcr, -1, axis=1)
    fvm = np.roll(fcr, -1, axis=0)
    fvp = np.roll(fcr, 1, axis=0)
    x0, y0 = np.meshgrid(np.arange(len(fcr)), np.arange(len(fcr)))

    summ1 = (2.0 * fcr - fhm - fhp)
    summ2 = (4.0 * fcr - 4.0 * fhm)
    summ3 = (8.0 * fcr + 4.0 * fhm - 2.0 * fhp + 4.0 * fvm - 2.0 * fvp)
    summ4 = (4.0 * fcr - 4.0 * fvm)
    summ5 = (2.0 * fcr - fvm - fvp)

    summ6 = (4.0 * fcr - 4.0 * fhp)
    summ7 = (10.0 * fcr - 2.0 * fhm + 4.0 * fhp - fvm + fvp)
    summ8 = (20.0 * fcr - 4.0 * fhm + 8.0 * fhp)

    summ9 = (8.0 * fcr - 2.0 * fhm + 4.0 * fhp - 2.0 * fvm + 4.0 * fvp)
    summ10 = (4.0 * fcr - 4.0 * fvp)
    summ11 = (2.0 * fcr - fvm - fvp)

    a, b = ha2, hb2
    a, b = - 1.0 / a, len(fcr) + b / a
    x1 = (-b + y0) / a
    x2 = (1 - b + y0) / a

    formula = a * x0 + b - y0
    formula_2 = formula * formula
    formula_3 = formula_2 * formula
    formula_4 = formula_3 * formula

    case1 = (
        + (1.0 / (24.0 * (a ** 3))) * (
            + summ1 * formula_4
            + a * summ2 * formula_3
            + (a ** 2) * (- summ3 * formula_2 - summ4 * formula_3 + summ5 * formula_4)
        )
    )

    formula = a + a * x0 + b - y0
    formula_2 = formula * formula
    formula_3 = formula_2 * formula

    case2 = (
        fcr + (1.0 / (24.0 * (a ** 3))) * (
            + 4.0 * summ1 * (- 0.25 + formula - 1.5 * formula_2 + formula_3)
            + (a * 3.0) * summ6 * (-1.0 / 3 + formula - formula_2)
            + (a ** 2) * (summ7 - summ8 * formula)
        )
    )

    formula = - 1.0 + a + a * x0 + b - y0
    formula_2 = formula * formula
    formula_3 = formula_2 * formula
    formula_4 = formula_3 * formula

    case3 = (
        fcr + (1.0 / (24.0 * (a ** 3))) * (
            - summ1 * formula_4
            + a * summ6 * formula_3
            + (a ** 2) * (summ9 * formula_2 - summ10 * formula_3 - summ11 * formula_4)
        )
    )

    new_data = np.where((x1 < x0) & (x2 < x0), 0, new_data)
    new_data = np.where((x1 > x0) & (x2 < x0), case1, new_data)
    new_data = np.where((x1 > x0) & (x0 + 1 > x1) & (x2 > x0) & (x0 + 1 > x2), case2, new_data)
    new_data = np.where((x1 > x0 + 1) & (x2 < x0 + 1), case3, new_data)

    # lower edge

    a, b = ha1, hb1
    a, b = - 1.0 / a, len(fcr) + b / a
    x1 = (-b + y0) / a
    x2 = (1 - b + y0) / a

    formula = a * x0 + b - y0
    formula_2 = formula * formula
    formula_3 = formula_2 * formula
    formula_4 = formula_3 * formula

    case1 = (
        + fcr - (1.0 / (24.0 * (a ** 3))) * (
            + summ1 * formula_4
            + a * summ2 * formula_3
            + (a ** 2) * (- summ3 * formula_2 - summ4 * formula_3 + summ5 * formula_4)
        )
    )

    formula = a + a * x0 + b - y0
    formula_2 = formula * formula
    formula_3 = formula_2 * formula

    case2 = (
        - (1.0 / (24.0 * (a ** 3))) * (
            + 4.0 * summ1 * (- 0.25 + formula - 1.5 * formula_2 + formula_3)
            + (a * 3.0) * summ6 * (-1.0 / 3 + formula - formula_2)
            + (a ** 2) * (summ7 - summ8 * formula)
        )
    )

    formula = - 1.0 + a + a * x0 + b - y0
    formula_2 = formula * formula
    formula_3 = formula_2 * formula
    formula_4 = formula_3 * formula

    case3 = (
        - (1.0 / (24.0 * (a ** 3))) * (
            - summ1 * formula_4
            + a * summ6 * formula_3
            + (a ** 2) * (summ9 * formula_2 - summ10 * formula_3 - summ11 * formula_4)
        )
    )

    new_data = np.where((x1 > x0) & (x2 < x0), case1, new_data)
    new_data = np.where((x1 > x0) & (x0 + 1 > x1) & (x2 > x0) & (x0 + 1 > x2), case2, new_data)
    new_data = np.where((x1 > x0 + 1) & (x2 < x0 + 1), case3, new_data)
    new_data = np.where((x1 > x0 + 1) & (x2 > x0 + 1), 0, new_data)

    new_data = np.rot90(new_data, 3)

    # error array

    xx = np.where(fits[1].data == 0, 1, fits[1].data)
    error = np.sqrt(new_data / xx) * fits[2].data

    flux = np.sum(new_data)
    error = np.sqrt(np.nansum(error * error))

    return flux, error


def get_flux_gauss(fits, lower_wavelength, upper_wavelength,
                   aperture_lower_extend, aperture_upper_extend, sigma, plot=False):

    spectrum_direction = variables.spectrum_direction.custom_from_fits(fits).value
    scan_length = variables.scan_length.custom_from_fits(fits).value
    scan_frame = variables.scan_frame.custom_from_fits(fits).value
    wavelength_frame = variables.wavelength_frame.custom_from_fits(fits).value

    if spectrum_direction > 0:
        y1 = min(aperture_lower_extend, aperture_upper_extend)
        y2 = scan_length + max(aperture_lower_extend, aperture_upper_extend)
    else:
        y1 = - scan_length - max(aperture_lower_extend, aperture_upper_extend)
        y2 = - min(aperture_lower_extend, aperture_upper_extend)

    science_frame = np.array(fits[plc.fits_sci(fits)[0]].data)
    error_frame = np.array(fits[plc.fits_err(fits)[0]].data)
    ph_error_frame = np.sqrt(np.abs(science_frame))

    scan_weight = (scipy.special.erf((scan_frame - y1) / ((sigma / 45.) * np.sqrt(2.0))) -
                   scipy.special.erf((scan_frame - y2) / ((sigma / 45.) * np.sqrt(2.0)))) / 2

    wavelength_weight = (scipy.special.erf((wavelength_frame - lower_wavelength) / (sigma * np.sqrt(2.0))) -
                         scipy.special.erf((wavelength_frame - upper_wavelength) / (sigma * np.sqrt(2.0)))) / 2

    weighted_science_frame = science_frame * scan_weight * wavelength_weight
    weighted_error_frame = error_frame * scan_weight * wavelength_weight
    weighted_ph_error_frame = ph_error_frame * scan_weight * wavelength_weight

    flux = np.sum(weighted_science_frame)
    error = np.sqrt(np.nansum(weighted_error_frame * weighted_error_frame))
    ph_error = np.sqrt(np.nansum(weighted_ph_error_frame * weighted_ph_error_frame))

    if plot:
        get_flux_integral(fits, lower_wavelength, upper_wavelength,
                          aperture_lower_extend, aperture_upper_extend, sigma, plot=True)

    return flux, error, ph_error


def photometry(input_data, white_lower_wavelength=None, white_upper_wavelength=None, bins_file=None,
               aperture_lower_extend=None, aperture_upper_extend=None, extraction_method=None,
               extraction_gauss_sigma=None, plot=False):

    # load pipeline and calibration variables to be used

    white_lower_wavelength = variables.white_lower_wavelength.custom(white_lower_wavelength)
    white_upper_wavelength = variables.white_upper_wavelength.custom(white_upper_wavelength)
    bins_file = variables.bins_file.custom(bins_file)
    aperture_lower_extend = variables.aperture_lower_extend.custom(aperture_lower_extend)
    aperture_upper_extend = variables.aperture_upper_extend.custom(aperture_upper_extend)
    extraction_method = variables.extraction_method.custom(extraction_method)
    extraction_gauss_sigma = variables.extraction_gauss_sigma.custom(extraction_gauss_sigma)

    ra_target = variables.ra_target.custom()
    dec_target = variables.dec_target.custom()
    subarray_size = variables.sub_array_size.custom()
    grism = variables.grism.custom()
    exposure_time = variables.exposure_time.custom()
    bins_number = variables.bins_number.custom()
    heliocentric_julian_date = variables.heliocentric_julian_date.custom()
    spectrum_direction = variables.spectrum_direction.custom()
    sky_background_level = variables.sky_background_level.custom()
    y_star = variables.y_star.custom()
    y_shift_error = variables.y_shift_error.custom()
    x_star = variables.x_star.custom()
    x_shift_error = variables.x_shift_error.custom()
    scan_length = variables.scan_length.custom()
    scan_length_error = variables.scan_length_error.custom()
    heliocentric_julian_date_array = variables.heliocentric_julian_date_array.custom()
    spectrum_direction_array = variables.spectrum_direction_array.custom()
    sky_background_level_array = variables.sky_background_level_array.custom()
    x_star_array = variables.x_star_array.custom()
    x_shift_error_array = variables.x_shift_error_array.custom()
    y_star_array = variables.y_star_array.custom()
    y_shift_error_array = variables.y_shift_error_array.custom()
    scan_length_array = variables.scan_length_array.custom()
    scan_length_error_array = variables.scan_length_error_array.custom()
    white_ldc1 = variables.white_ldc1.custom()
    white_ldc2 = variables.white_ldc2.custom()
    white_ldc3 = variables.white_ldc3.custom()
    white_ldc4 = variables.white_ldc4.custom()

    lower_wavelength = variables.lower_wavelength.custom()
    upper_wavelength = variables.upper_wavelength.custom()
    flux_array = variables.flux_array.custom()
    error_array = variables.error_array.custom()
    ph_error_array = variables.ph_error_array.custom()

    # set bins

    white_dictionary, bins_dictionaries = \
        variables.set_binning(input_data, white_lower_wavelength.value, white_upper_wavelength.value,
                              white_ldc1.value, white_ldc2.value, white_ldc3.value, white_ldc4.value,
                              bins_file.value)

    # select extraction method

    used_extraction_method = {'integral': get_flux_integral, 'gauss': get_flux_gauss}[extraction_method.value]

    # initiate counter

    counter = PipelineCounter('Photometry', len(input_data.spectroscopic_images))

    # iterate over the list of HDUList objects included in the input data

    light_curve = {}

    for fits in input_data.spectroscopic_images:

        try:
            ra_target.from_dictionary(light_curve)

        except KeyError:

            ra_target.from_fits(fits)
            ra_target.to_dictionary(light_curve)

            dec_target.from_fits(fits)
            dec_target.to_dictionary(light_curve)

            subarray_size.set(len(fits[1].data))
            subarray_size.to_dictionary(light_curve)

            grism.from_fits(fits)
            grism.to_dictionary(light_curve)

            exposure_time.from_fits(fits)
            exposure_time.to_dictionary(light_curve)

            aperture_lower_extend.to_dictionary(light_curve)

            aperture_upper_extend.to_dictionary(light_curve)

            extraction_method.to_dictionary(light_curve)

            extraction_gauss_sigma.to_dictionary(light_curve)

        heliocentric_julian_date.from_fits(fits)
        heliocentric_julian_date_array.set(
            np.append(heliocentric_julian_date_array.value, heliocentric_julian_date.value))

        spectrum_direction.from_fits(fits)
        spectrum_direction_array.set(np.append(spectrum_direction_array.value, spectrum_direction.value))

        sky_background_level.from_fits(fits, position=plc.fits_sci(fits)[0])
        sky_background_level_array.set(np.append(sky_background_level_array.value, sky_background_level.value))

        y_star.from_fits(fits)
        y_star_array.set(np.append(y_star_array.value, y_star.value))

        y_shift_error.from_fits(fits)
        y_shift_error_array.set(np.append(y_shift_error_array.value, y_shift_error.value))

        x_star.from_fits(fits)
        x_star_array.set(np.append(x_star_array.value, x_star.value))

        x_shift_error.from_fits(fits)
        x_shift_error_array.set(np.append(x_shift_error_array.value, x_shift_error.value))

        scan_length.from_fits(fits)
        scan_length_array.set(np.append(scan_length_array.value, scan_length.value))

        scan_length_error.from_fits(fits)
        scan_length_error_array.set(np.append(scan_length_error_array.value, scan_length_error.value))

        bins_number.set(len(bins_dictionaries))
        bins_number.to_dictionary(light_curve)

        for i in [white_dictionary] + bins_dictionaries:
            lower_wavelength.from_dictionary(i)
            upper_wavelength.from_dictionary(i)
            flux, error, ph_error = used_extraction_method(fits, lower_wavelength.value, upper_wavelength.value,
                                                           aperture_lower_extend.value, aperture_upper_extend.value,
                                                           extraction_gauss_sigma.value)
            flux_array.from_dictionary(i)
            flux_array.to_dictionary(i, value=np.append(flux_array.value, flux))
            error_array.from_dictionary(i)
            error_array.to_dictionary(i, value=np.append(error_array.value, error))
            ph_error_array.from_dictionary(i)
            ph_error_array.to_dictionary(i, value=np.append(ph_error_array.value, ph_error))

        counter.update()

        if plot:

            plt.figure(1)
            plt.imshow(fits[1].data, origin='lower', aspect='auto')
            plt.xlim(0, len(fits[1].data))
            plt.ylim(0, len(fits[1].data))

            lower_wavelength.from_dictionary(white_dictionary)
            upper_wavelength.from_dictionary(white_dictionary)

            used_extraction_method(fits, lower_wavelength.value, upper_wavelength.value,
                                   aperture_lower_extend.value, aperture_upper_extend.value,
                                   extraction_gauss_sigma.value, plot=True)

            plt.xlabel(r'$\mathrm{column \, (pix)}$', fontsize=20)
            plt.ylabel(r'$\mathrm{row \, (pix)}$', fontsize=20)

            plt.figure(2)

            plot_bins = np.arange(10000, 18000, 50)
            if grism.value == 'G102':
                plot_bins = np.arange(6000, 13000, 50)
            plot_spectrum = np.array([used_extraction_method(fits, ff, ff + 50, aperture_lower_extend.value,
                                                             aperture_upper_extend.value,
                                                             extraction_gauss_sigma.value)[0]
                                      for ff in plot_bins])

            plt.plot((plot_bins + 25) / 10000.0, plot_spectrum / 1000000.0, 'r-', lw=2)

            plt.ylabel(r'$\mathrm{e}^{-} \, (\times 10^6)$', fontsize=20)
            plt.xlabel(r'$\lambda \, [\mu \mathrm{m}]$', fontsize=20)

    for i in [heliocentric_julian_date_array, spectrum_direction_array, sky_background_level_array, x_star_array,
              x_shift_error_array, y_star_array, y_shift_error_array, scan_length_array, scan_length_error_array,
              white_dictionary] + bins_dictionaries:
        i.to_dictionary(light_curve)

    if plot:
        return input_data, [plt.figure(1), plt.figure(2)]

    else:
        return input_data, light_curve


def split_photometry(input_data, white_lower_wavelength=None, white_upper_wavelength=None, bins_file=None,
               aperture_lower_extend=None, aperture_upper_extend=None, extraction_method=None,
               extraction_gauss_sigma=None, plot=False):

    # load pipeline and calibration variables to be used

    white_lower_wavelength = variables.white_lower_wavelength.custom(white_lower_wavelength)
    white_upper_wavelength = variables.white_upper_wavelength.custom(white_upper_wavelength)
    bins_file = variables.bins_file.custom(bins_file)
    aperture_lower_extend = variables.aperture_lower_extend.custom(aperture_lower_extend)
    aperture_upper_extend = variables.aperture_upper_extend.custom(aperture_upper_extend)
    extraction_method = variables.extraction_method.custom(extraction_method)
    extraction_gauss_sigma = variables.extraction_gauss_sigma.custom(extraction_gauss_sigma)

    ra_target = variables.ra_target.custom()
    dec_target = variables.dec_target.custom()
    subarray_size = variables.sub_array_size.custom()
    grism = variables.grism.custom()
    exposure_time = variables.exposure_time.custom()
    bins_number = variables.bins_number.custom()
    heliocentric_julian_date_array = variables.heliocentric_julian_date_array.custom()
    spectrum_direction_array = variables.spectrum_direction_array.custom()
    sky_background_level_array = variables.sky_background_level_array.custom()
    x_star_array = variables.x_star_array.custom()
    x_shift_error_array = variables.x_shift_error_array.custom()
    y_star_array = variables.y_star_array.custom()
    y_shift_error_array = variables.y_shift_error_array.custom()
    scan_length_array = variables.scan_length_array.custom()
    scan_length_error_array = variables.scan_length_error_array.custom()
    white_ldc1 = variables.white_ldc1.custom()
    white_ldc2 = variables.white_ldc2.custom()
    white_ldc3 = variables.white_ldc3.custom()
    white_ldc4 = variables.white_ldc4.custom()

    lower_wavelength = variables.lower_wavelength.custom()
    upper_wavelength = variables.upper_wavelength.custom()
    flux_array = variables.flux_array.custom()
    error_array = variables.error_array.custom()
    ph_error_array = variables.ph_error_array.custom()

    # set bins

    white_dictionary, bins_dictionaries = \
        variables.set_binning(input_data, white_lower_wavelength.value, white_upper_wavelength.value,
                              white_ldc1.value, white_ldc2.value, white_ldc3.value, white_ldc4.value,
                              bins_file.value)

    # iterate over the splitted data sub-sets

    final_light_curve = {}

    for split_number, splitted_sub_set in enumerate(input_data.spectroscopic_images):

        if not plot:
            print('Splitting sample {0}:'.format(split_number + 1))

        light_curve = \
            photometry(input_data.copy_split(split_number),
                       white_lower_wavelength=white_lower_wavelength.value,
                       white_upper_wavelength=white_upper_wavelength.value,
                       aperture_lower_extend=aperture_lower_extend.value,
                       aperture_upper_extend=aperture_upper_extend.value,
                       bins_file=bins_file.value,
                       extraction_method=extraction_method.value,
                       extraction_gauss_sigma=extraction_gauss_sigma.value,
                       plot=False)[1]

        final_light_curve[variables.light_curve_split.keyword + str(split_number + 1)] = light_curve

        if plot:

            fits = splitted_sub_set[0]
            used_extraction_method = {'integral': get_flux_integral, 'gauss': get_flux_gauss}[extraction_method.value]

            total_plots = len(input_data.spectroscopic_images)
            plot_columns = 3
            plot_rows = int(total_plots / float(plot_columns) - 0.0000000001) + 1

            plt.figure(1, figsize=(3 * plot_columns, 3 * plot_rows))
            plt.subplot(plot_rows, plot_columns, split_number + 1)
            plt.title(r'{0}{1}{2}'.format('$\mathrm{split \, ', str(split_number + 1), '}$'), fontsize=20)
            if split_number + 1 != plot_columns * (plot_rows - 1) + 1:
                plt.tick_params(labelleft=False, labelbottom=False)

            plt.imshow(fits[1].data, origin='lower', aspect='auto')
            plt.xlim(0, len(fits[1].data))
            plt.ylim(0, len(fits[1].data))

            lower_wavelength.from_dictionary(white_dictionary)
            upper_wavelength.from_dictionary(white_dictionary)

            used_extraction_method(fits, lower_wavelength.value, upper_wavelength.value,
                                   aperture_lower_extend.value, aperture_upper_extend.value,
                                   extraction_gauss_sigma.value, plot=True)

            plt.figure(2)
            testx = np.arange(10000, 18000, 50)
            if grism.value == 'G102':
                testx = np.arange(6000, 13000, 50)
            testy = np.array([used_extraction_method(fits, ff, ff + 50, aperture_lower_extend.value,
                                                     aperture_upper_extend.value, extraction_gauss_sigma.value)[0]
                              for ff in testx])

            plt.plot((testx + 25) / 10000.0, testy / 1000000.0, '-', lw=2,
                     label=r'{0}{1}{2}'.format('$\mathrm{split \, ', str(split_number + 1), '}$'))

            if split_number + 1 == total_plots:

                plt.figure(1).text(0.005, 0.5, r'$\mathrm{row \, (pix)}$', fontsize=20,
                                   ha='center', va='center', rotation='vertical')
                plt.figure(1).text(0.5, 0.01, r'$\mathrm{column \, (pix)}$', fontsize=20, ha='center', va='center')

                plt.figure(2)
                plt.xlim(0.9, 2.0)
                plt.legend()
                plt.ylabel(r'$\mathrm{e}^{-} \, (\times 10^6)$', fontsize=20)
                plt.xlabel(r'$\lambda \, [\mu \mathrm{m}]$', fontsize=20)

        if split_number == 0:

            ra_target.from_dictionary(light_curve)
            ra_target.to_dictionary(final_light_curve)

            dec_target.from_dictionary(light_curve)
            dec_target.to_dictionary(final_light_curve)

            subarray_size.from_dictionary(light_curve)
            subarray_size.to_dictionary(final_light_curve)

            grism.from_dictionary(light_curve)
            grism.to_dictionary(final_light_curve)

            exposure_time.from_dictionary(light_curve)
            exposure_time.to_dictionary(final_light_curve)

            bins_number.from_dictionary(light_curve)
            bins_number.to_dictionary(final_light_curve)

            aperture_lower_extend.from_dictionary(light_curve)
            aperture_lower_extend.to_dictionary(final_light_curve)

            aperture_upper_extend.from_dictionary(light_curve)
            aperture_upper_extend.to_dictionary(final_light_curve)

            extraction_method.from_dictionary(light_curve)
            extraction_method.to_dictionary(final_light_curve)

            extraction_gauss_sigma.from_dictionary(light_curve)
            extraction_gauss_sigma.to_dictionary(final_light_curve)

            heliocentric_julian_date_array.from_dictionary(light_curve)
            heliocentric_julian_date_array.to_dictionary(final_light_curve)

            spectrum_direction_array.from_dictionary(light_curve)
            spectrum_direction_array.to_dictionary(final_light_curve)

            sky_background_level_array.from_dictionary(light_curve)
            sky_background_level_array.to_dictionary(final_light_curve)

            x_star_array.from_dictionary(light_curve)
            x_star_array.to_dictionary(final_light_curve)

            x_shift_error_array.from_dictionary(light_curve)
            x_shift_error_array.to_dictionary(final_light_curve)

            y_star_array.from_dictionary(light_curve)
            y_star_array.to_dictionary(final_light_curve)

            y_shift_error_array.from_dictionary(light_curve)
            y_shift_error_array.to_dictionary(final_light_curve)

            scan_length_array.from_dictionary(light_curve)
            scan_length_array.to_dictionary(final_light_curve)

            scan_length_error_array.from_dictionary(light_curve)
            scan_length_error_array.to_dictionary(final_light_curve)

            for i in [white_dictionary] + bins_dictionaries:
                i.from_dictionary(light_curve)
                i.to_dictionary(final_light_curve)

        else:

            for i in [white_dictionary] + bins_dictionaries:
                i.from_dictionary(light_curve)
                flux_array.from_dictionary(i)
                error_array.from_dictionary(i)
                ph_error_array.from_dictionary(i)
                current_flux = flux_array.value
                current_error = error_array.value
                current_ph_error = ph_error_array.value
                i.from_dictionary(final_light_curve)
                flux_array.from_dictionary(i)
                flux_array.to_dictionary(i, value=flux_array.value + current_flux)
                error_array.from_dictionary(i)
                error_array.to_dictionary(i, value=np.sqrt(error_array.value ** 2 + current_error ** 2))
                ph_error_array.from_dictionary(i)
                ph_error_array.to_dictionary(i, value=np.sqrt(ph_error_array.value ** 2 + current_ph_error ** 2))
                i.to_dictionary(final_light_curve)

    if plot:

        return input_data, [plt.figure(1), plt.figure(2)]

    else:

        return input_data, final_light_curve


def plot_photometry(dataset, lightcurve, directory):

    forward_colour = 'k'
    reverse_colour = 'r'

    hjd_time = lightcurve[variables.heliocentric_julian_date_array.keyword]
    flux = lightcurve[variables.white_dictionary.keyword][variables.flux_array.keyword]
    ssky = lightcurve[variables.sky_background_level_array.keyword]
    scan = lightcurve[variables.spectrum_direction_array.keyword]
    reverse = np.where(np.array(scan) < 0)
    forward = np.where(np.array(scan) > 0)

    if len(forward[0]) > 0:
        if dataset.splitted:
            original = list(dataset.spectroscopic_images)
            dataset.spectroscopic_images = list(np.array(dataset.spectroscopic_images)[:, forward[0]])
            test = []
            for i in dataset.spectroscopic_images:
                test.append([i[-1]])
            dataset.spectroscopic_images = test
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                figures = split_photometry(dataset, plot=True)[1]
            dataset.spectroscopic_images = original
        else:
            original = list(dataset.spectroscopic_images)
            dataset.spectroscopic_images = [dataset.spectroscopic_images[forward[0][-1]]]
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                figures = photometry(dataset, plot=True)[1]
            dataset.spectroscopic_images = original

        plc.save_figure(directory, figure=figures[0], name='forward_extraction_aperture')
        plc.save_figure(directory, figure=figures[1], name='forward_stellar_spectrum')

        plt.close('all')

    if len(reverse[0]) > 0:
        if dataset.splitted:
            dataset.spectroscopic_images = list(np.array(dataset.spectroscopic_images)[:, reverse[0]])
            test = []
            for i in dataset.spectroscopic_images:
                test.append([i[-1]])
            dataset.spectroscopic_images = test
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                figures = split_photometry(dataset, plot=True)[1]
        else:
            dataset.spectroscopic_images = [dataset.spectroscopic_images[reverse[0][-1]]]
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                figures = photometry(dataset, plot=True)[1]

        plc.save_figure(directory, figure=figures[0], name='reverse_extraction_aperture')
        plc.save_figure(directory, figure=figures[1], name='reverse_stellar_spectrum')

        plt.close('all')

    plt.subplot(2, 1, 1)
    plt.plot((np.array(hjd_time) - hjd_time[0])[forward], np.array(flux)[forward] / (10 ** 8),
             'o', c=forward_colour, mec=forward_colour, ms=3)
    plt.plot((np.array(hjd_time) - hjd_time[0])[reverse], np.array(flux)[reverse] / (10 ** 8),
             'o', c=reverse_colour, mec=reverse_colour, ms=3)
    plt.ylabel(r'$\mathrm{e}^{-} \, (\times 10^8)$', fontsize=15)
    plt.tick_params(labelbottom=False)
    plc.adjust_ticks()
    plt.subplot(2, 1, 2)
    plt.plot((np.array(hjd_time) - hjd_time[0])[forward], np.array(ssky)[forward],
             'o', c=forward_colour, mec=forward_colour, ms=3)
    plt.plot((np.array(hjd_time) - hjd_time[0])[reverse], np.array(ssky)[reverse],
             'o', c=reverse_colour, mec=reverse_colour, ms=3)
    plt.xlabel(r'$\Delta t \, \mathrm{(days)}$', fontsize=15)
    plt.ylabel(r'$\mathrm{sky} \, \mathrm{ratio}$', fontsize=15)
    plc.adjust_ticks()
    plt.subplots_adjust(hspace=0)
    plc.save_figure(directory, name='raw_light_curve')



"""
File: images1_reduction.py
Author: Angelos Tsiaras
E-mail: angelos.tsiaras.14@ucl.ac.uk
Contents: functions that perform reduction processes
    timing:     Heliocentric julian date calculation.
    bias:       Bias drifts and zero-read corrections.
    linearity:
    dark:
    gain:       Gain variations correction and DN to electrons conversion.
    sky:
    flat:
    bpcr:

"""

__all__ = ['timing', 'bias', 'dark', 'flat', 'linearity', 'gain', 'sky', 'bpcr']

import numpy as np
import warnings
import pylightcurve as plc

from scipy.interpolate import griddata
from scipy.interpolate import interp1d

from iraclis.classes import *


def timing(input_data):
    """
    Heliocentric julian date calculation.

    Calculates the heliocentric julian date of the mid-exposure (hjd):

    .. math:: hjd = jd - (d / c) (\sin(d) \sin(ds_{jd}) + \cos(d) \cos(ds_{jd}) \cos(a - as_{jd})

    with:

    .. math:: jd = 0.5 (exp_s + exp_e) + 2400000.5

    where:

    jd is the julian date of the mid-exposure,
    d is the distance between the earth and the sun, 149597870700.0 km,
    c is the speed of light, 25902068371200.0 km/day,
    exp_s and exp_e are the exposure start and end times in modified julian date (included in the data file),
    a and d are the target RA and DEC (included in the data file) and
    as_jd and ds_jd are the sun RA and DEC at the julian date of the mid-exposure.

    Parameters
    ----------
    input_data : HDUList or DataSet object
        input data, either a single image or a complete data set

    Returns
    -------
    input_data : HDUList or DataSet object
        updated version of the input_data object, including the heliocentric julian date of the mid-exposure

    """

    # load pipeline and calibration variables to be used

    exposure_start = variables.exposure_start.custom()
    exposure_end = variables.exposure_end.custom()
    ra_target = variables.ra_target.custom()
    dec_target = variables.dec_target.custom()
    heliocentric_julian_date = variables.heliocentric_julian_date.custom()

    # initiate counter

    counter = PipelineCounter('Timing', len(input_data.spectroscopic_images))

    # iterate over the list of HDUList objects included in the input data

    for fits in input_data.spectroscopic_images:

        # get the exposure start and end times, as well as the target RA and DEC from the fits file

        exposure_start.from_fits(fits)
        exposure_end.from_fits(fits)
        ra_target.from_fits(fits)
        dec_target.from_fits(fits)

        # calculate the heliocentric julian date of the mid-exposure

        heliocentric_julian_date.to_fits(fits, value=plc.mjd_to_hjd(ra_target.value, dec_target.value,
                                                                    0.5 * (exposure_start.value + exposure_end.value)))

        # update counter

        counter.update()

    # return the updated version of the input data

    return input_data


def bias(input_data):
    """
    Bias drifts and zero-read corrections.

    1) Corrects each NDR for the zero-read.

    As the WFC3/IR detector lacks a shutter, the pixels are exposed even before the exposure starts. Hence the ZR is a
    record of the light collected before the beginning of the exposure. Since this light is not part of the exposure,
    the ZR is subtracted from each consecutive NDR:

    .. math:: NDR \rightarrow NDR - ZR

    At this stage the first NDR is renamed from SCI, not to be considered as an NDR for the rest of the analysis.

    2) Corrects each NDR for the bias drifts.

    The reference pixels define a group of pixels that are not sensitive to incoming light and are located at the first
    five and last five rows and columns of the detector. If the bias level is varying between the zero-read and an NDR,
    the variation is imprinted in the reference pixels. For this reason, the mean value of the reference pixels
    (reference level, NDR_rl) is subtracted from each NDR. The first five and the last five rows, as well as the
    first and the last columns, are excluded from the calculation because they have been found to be unstable.

    .. math:: NDR \rightarrow NDR - NDR_{rl}

    3) Calculates the error-array of each NDR.

    This initial calculation of the error-array of an NDR (NDR_e), combines the detector read noise (RN) and the
    Poison noise. The detector read noise is included in the WFC3 ccd calibration file in electron units, while the
    Poison noise is the square root of the number of electrons recorded in an NDR. At this stage the NDR is in DN units,
    so the number of electrons is given by multiplying the NDR by the detector gain (G), which is also included in the
    WFC3 ccd calibration file. The final uncertainty is converted back to DN units by dividing by the detector gain.

    .. math:: NDR_e (e- units) = \sqrt( RN ^ 2 +  NDR * G )
    .. math:: NDR_e (DN units) = \sqrt( RN ^ 2 +  NDR * G ) / G

    The error-array of each NDR is stored in the data file just after the NDR.

    Parameters
    ----------
    input_data : HDUList or DataSet object
        input data, either a single image or a complete data set

    Returns
    -------
    input_data : HDUList or DataSet object
        updated version of the input_data object, corrected for bias drifts and zero-read

    """

    # load pipeline and calibration variables to be used

    zero_read_frame = variables.zero_read.custom()
    zero_read_error_frame = variables.zero_read_error.custom()
    reference_pixels_level = variables.reference_pixels_level.custom()
    
    ccd_gain = calibrations.ccd_gain.match(input_data)
    ccd_read_noise = calibrations.ccd_read_noise.match(input_data)

    # initiate counter

    counter = PipelineCounter('Bias', len(input_data.spectroscopic_images))

    # iterate over the list of HDUList objects included in the input data

    for fits in input_data.spectroscopic_images:

        # check if the zero-read has been identified

        try:

            # get the zero-read and the zero-read error-array from the fits file

            zero_read_frame.from_fits(fits)
            zero_read_error_frame.from_fits(fits)

        except KeyError:

            # locate and rename the zero-read and the zero-read error-array
            # (first non-destructive read - i.e. last in the list of scientific frames)

            fits[plc.fits_sci(fits)[-1]].name = zero_read_frame.keyword
            fits[plc.fits_err(fits)[-1]].name = zero_read_error_frame.keyword

            # get the zero-read and the zero-read error-array from the fits file

            zero_read_frame.from_fits(fits)
            zero_read_error_frame.from_fits(fits)

        # iterate over the NDRs

        for i, j in plc.fits_sci_err(fits):

            # get the NDR from the fits file

            science = np.array(fits[i].data, dtype=float)

            # correct the NDR for the the zero-read

            science -= zero_read_frame.value

            # calculate (and save in the data file) the reference level

            reference_pixels_level.to_fits(
                fits,
                value=np.mean(np.sort(np.insert(science[5:-5, -5:-1].flatten(), 0, science[5:-5, 1:5].flatten()))),
                position=i)
            
            # correct the NDR for the the bias drifts

            science -= reference_pixels_level.value

            # calculate the error-array of the NDR

            error = (np.sqrt(ccd_read_noise * ccd_read_noise + ccd_gain * np.abs(science)) / ccd_gain)

            # update the NDR and its error-array in the data file

            fits[i].data = science
            fits[j].data = error

        # update counter

        counter.update()

    # return the updated version of the input data

    return input_data


def linearity(input_data):
    """
    Non-linearity correction.

    1) Calculates the zero-read flux.

    The first NDR (zero-read, ZR) is a combination of the default bias level of the detector (super-zero-read, SZR) and
    the flux recorded before the beginning of the exposure (zero-read flux, ZRF). The super-zero-read is included in the
    WFC3 linearity calibration file. Given the above, the zero-read flux is calculated by subtracting from the zero-read
    the super-zero-read:

    .. math:: ZRF = ZR - SZR

    The ZRF is stored in the data file as the last element of the HDUList object.





    2) Calculates the error-array of the zero-read flux.



    .. math:: ZRF_e = \sqrt{RN ^ 2 + (SZR_e * G ) ^ 2 + ZRF * G} / G







    3) Corrects each read for the non-linearity behaviour of the IR detector, by applying on it the function:

    .. math:: F_{final} = F_c(f_r + f_z) - F_c(f_z)

    where fr is the flux in the read, fz is the zero-read flux, and Fc is the non-linearity correction function:

    .. math:: F_c(f) = (1 + c1 + c2 * f + c3 * f ^ 2 + c4 * f ^ 3) * f

    where c1, c2, c3, c4 are the non-linearity coefficient arrays included in the linearity calibration file.



    4)

    Also propagates the uncertainties in the error arrays.

    Parameters
    ----------
    input_data : HDUList or DataSet object
        input data, either a single image or a complete data set

    Returns
    -------
    input_data : HDUList or DataSet object
        updated version of the input_data object, corrected for non-linearity

    """

    # load pipeline and calibration variables to be used

    zero_read = variables.zero_read.custom()
    zero_read_error = variables.zero_read_error.custom()
    zero_read_flux = variables.zero_read_flux.custom()
    zero_read_flux_error = variables.zero_read_flux_error.custom()

    super_zero_read = calibrations.super_zero_read.match(input_data)
    super_zero_read_error = calibrations.super_zero_read_error.match(input_data)
    ccd_gain = calibrations.ccd_gain.match(input_data)
    ccd_read_noise = calibrations.ccd_read_noise.match(input_data)
    linearity_coefficient_1 = calibrations.linearity_coefficient_1.match(input_data)
    linearity_coefficient_2 = calibrations.linearity_coefficient_2.match(input_data)
    linearity_coefficient_3 = calibrations.linearity_coefficient_3.match(input_data)
    linearity_coefficient_4 = calibrations.linearity_coefficient_4.match(input_data)

    # initiate counter

    counter = PipelineCounter('Linearity', len(input_data.spectroscopic_images))

    # iterate over the list of HDUList objects included in the input data

    for fits in input_data.spectroscopic_images:

        # check if the zero-read has been identified

        try:

            # get the zero-read and the zero-read error-array from the fits file

            zero_read.from_fits(fits)
            zero_read_error.from_fits(fits)

        except KeyError:

            # locate and rename the zero-read and the zero-read error-array
            # (first non-destructive read - i.e. last in the list of scientific frames)

            fits[plc.fits_sci(fits)[-1]].name = zero_read.keyword
            fits[plc.fits_err(fits)[-1]].name = zero_read_error.keyword

            # get the zero-read and the zero-read error-array from the fits file

            zero_read.from_fits(fits)
            zero_read_error.from_fits(fits)

        # calculate (and save in the data file) the zero-read flux

        zero_read_flux.to_fits(fits, value=zero_read.value - super_zero_read)

        # calculate (and save in the data file) the zero-read flux error-array

        zero_read_flux_error.to_fits(fits, value=np.sqrt(np.abs((ccd_read_noise * ccd_read_noise +
                                                                 ccd_gain * np.abs(zero_read_flux.value))) /
                                                         (ccd_gain * ccd_gain) +
                                                         super_zero_read_error * super_zero_read_error))

        # iterate over the NDRs

        for i, j in plc.fits_sci_err(fits):

            # get the NDR and its error-array from the data file

            science = np.array(fits[i].data, dtype=float)
            error = np.array(fits[j].data, dtype=float)

            # just maths, to make the calculation shorter and faster

            c1 = linearity_coefficient_1
            c2 = linearity_coefficient_2
            c3 = linearity_coefficient_3
            c4 = linearity_coefficient_4

            f_r = science
            f_r2 = f_r * f_r

            f_z = zero_read_flux.value
            f_z2 = f_z * f_z
            f_z3 = f_z2 * f_z

            f_rpz = f_r + f_z
            f_rpz2 = f_rpz * f_rpz
            f_rpz3 = f_rpz2 * f_rpz

            f_rmz = f_r * f_z

            f_rzz = f_rpz + f_z

            f_r_err = error
            f_r_err2 = f_r_err * f_r_err

            f_z_err = zero_read_flux_error.value
            f_z_err2 = f_z_err * f_z_err

            # apply the non-linearity correction function

            science = (
                f_rpz * (1.0 + c1 + c2 * f_rpz + c3 * f_rpz2 + c4 * f_rpz3) -
                f_z * (1.0 + c1 + c2 * f_z + c3 * f_z2 + c4 * f_z3)
            )

            # propagate the uncertainties in the error-array

            error = np.sqrt(
                f_r_err2 * ((1.0 + c1 + f_rpz * (2.0 * c2 + f_rpz * (3.0 * c3 + 4.0 * c4 * f_rpz))) ** 2) -
                f_z_err2 * f_r2 * ((2.0 * c2 + 3.0 * c3 * f_rzz + 4.0 * c4 * (f_r2 + 3.0 * f_rmz + 3.0 * f_z2)) ** 2)
            )

            # update the NDR and its error-array in the data file

            fits[i].data = science
            fits[j].data = error

        # update counter

        counter.update()

    # return the updated version of the input data

    return input_data


def dark(input_data, splitting=False):
    """
    Dark current correction.

    1) Corrects each NDR for the dark current.

    by subtracting from it the suitable dark current frame
    included in the suitable dark calibration file.
    This frame is selected in order to have the same sub-array, sampling sequence and sample number.

    2) Propagates the uncertainties in the error-array of each NDR.

    The existing error-array is combined with the error-array of the selected dark current (DC_e).

    .. math:: NDR_e \rightarrow \sqrt{NDR_e ^ 2 + DC_e ^ 2}

    Parameters
    ----------
    input_data : HDUList or DataSet object
        input data, either a single image or a complete data set

    splitting : bool
        whether to apply dark current correction on splitted data or not

    Returns
    -------
    input_data : HDUList or DataSet object
        updated version of the input_data object, corrected for dark current

    """

    # load pipeline and calibration variables to be used

    super_dark = calibrations.super_dark.match(input_data)

    # initiate counter

    counter = PipelineCounter('Dark', len(input_data.spectroscopic_images))

    # iterate over the list of HDUList objects included in the input data

    for fits in input_data.spectroscopic_images:

        # iterate over the NDRs

        for i, j in plc.fits_sci_err(fits):

            # get the NDR and its error-array from the data file

            science = np.array(fits[i].data, dtype=float)
            error = np.array(fits[j].data, dtype=float)

            if splitting:
                dark_current = np.array(super_dark[i].data - super_dark[i + 5].data, dtype=float)
                dark_current_error = np.sqrt(np.array(super_dark[j].data ** 2 + super_dark[j + 5].data ** 2,
                                                      dtype=float))
            else:
                dark_current = np.array(super_dark[i].data, dtype=float)
                dark_current_error = np.array(super_dark[j].data, dtype=float)

            # subtract the dark current

            science -= dark_current

            # propagate the uncertainties in the error-array

            error = np.sqrt(error * error + dark_current_error * dark_current_error)

            # update the NDR and its error-array in the data file

            fits[i].data = science
            fits[j].data = error

        # update counter

        counter.update()

    # return the updated version of the input data

    return input_data


def gain(input_data):
    """
    Gain variations correction and DN to electrons conversion.

    1) Corrects each NDR for the gain variations among the quarters.

    The WFC3/IR detector consists of four different parts (quadrants) which have different A/D converters. To correct
    for the differences in their gain each NDR is divided by the detector gain pattern (GP), which is included in the
    WFC3 pfl calibration file, and gives the ratio between the gain of each pixel and the average gain of the whole
    detector.

    .. math:: NDR \rightarrow NDR / GP
       
    2) Converts each NDR from DN units to electron units.

    For this correction, each NDR is multiplied by the mean gain of the whole detector (MG), which is included in the
    WFC3 ccd calibration file, and is the average electrons / DN rate of the whole detector.

    .. math:: NDR \rightarrow NDR * MG

    3) Converts the error-array of each NDR from DN units to electron units.

    The same conversion that is applied to each NDR is applied to the each NDR error-array (NDR_e) as well.

    .. math:: NDR_e \rightarrow NDR_e * MG / GP

    Parameters
    ----------
    input_data : HDUList or DataSet object
        input data, either a single image or a complete data set

    Returns
    -------
    input_data : HDUList or DataSet object
        updated version of the input_data object, corrected for gain variations and converted from DN to electrons

    """

    # load pipeline and calibration variables to be used

    gain_profile = calibrations.gain_profile.match(input_data)
    mean_gain = calibrations.mean_gain.match(input_data)

    # initiate counter

    counter = PipelineCounter('Gain', len(input_data.spectroscopic_images))

    # iterate over the list of HDUList objects included in the input data

    for fits in input_data.spectroscopic_images:

        # iterate over the NDRs

        for i, j in plc.fits_sci_err(fits):

            # get the NDR and its error-array from the data file

            science = np.array(fits[i].data, dtype=float)
            error = np.array(fits[j].data, dtype=float)

            # correct the NDR for the gain variations

            science /= gain_profile

            # convert the NDR from DN units to electron units

            science *= mean_gain

            # convert the error-array of the NDR from DN units to electron units

            error *= mean_gain / gain_profile

            # update the NDR and its error-array in the data file

            fits[i].data = science
            fits[j].data = error

        # update counter

        counter.update()

    # return the updated version of the input data

    return input_data


def sky(input_data, sky_detection_limit=None, splitting=False):
    """
    Sky background correction.

    Corrects each read for the sky background, by subtracting a scaled version of the master sky frame
    included in the suitable sky calibration file.

    The sky calibration file is selected in order to have the same grism (G102 or G141). The scaling factor
    (or sky ratio) is the median of the division between the read and the master sky frame,
    as calculated from a non-illuminated area of the detector.

    The non-illuminated pixels are found based on all the possible differential reads.
    A non-illuminated pixel has to be within mean +/- x * sigma (x = sky_detection_limit)
    in all the possible differential reads, after an eight-pixels moving average soothing
    (this is not affecting the science frame, it is performed only for the detection of the non-illuminated pixels).

    Also propagates the uncertainties in the error arrays.

    Parameters
    ----------
    input_data : HDUList or DataSet object
        input data, either a single image or a complete data set

    sky_detection_limit : float
        detection threshold for the non-illuminated pixels

    splitting : bool
        whether to apply dark current correction on splitted data or not

    Returns
    -------
    input_data : HDUList or DataSet object
        updated version of the input_data object, corrected for sky background

    """

    # load pipeline and calibration variables to be used

    sky_detection_limit = variables.sky_detection_limit.custom(sky_detection_limit)

    sky_background_level = variables.sky_background_level.custom()
    sky_frame = variables.sky_area.custom()

    master_sky = calibrations.master_sky.match(input_data)

    # filter the zero values in the master-sky

    master_sky = np.where(master_sky == 0, 1, master_sky)

    # initiate counter

    counter = PipelineCounter('Sky', len(input_data.spectroscopic_images))

    # iterate over the list of HDUList objects included in the input data

    for fits in input_data.spectroscopic_images:

        if not splitting:

            # calculate all the differential combinations of the NDRs (the first read is included as it is)

            differential_science = [fits[plc.fits_sci(fits)[-1]].data]

            for i in range(len(plc.fits_sci(fits))):
                for j in range(i + 1, len(plc.fits_sci(fits))):
                    differential_science.append(np.array(fits[plc.fits_sci(fits)[i]].data -
                                                         fits[plc.fits_sci(fits)[j]].data))

            # apply an eight-pixel moving average smoothing on the differential combinations

            for i in range(len(differential_science)):
                frame = differential_science[i]
                differential_science[i] = \
                    np.mean([frame,
                             np.roll(frame, 1, 0), np.roll(frame, -1, 0), np.roll(frame, 1, 1), np.roll(frame, -1, 1),
                             np.roll(np.roll(frame, 1, 0), 1, 1), np.roll(np.roll(frame, 1, 0), -1, 1),
                             np.roll(np.roll(frame, -1, 0), 1, 1), np.roll(np.roll(frame, -1, 0), -1, 1)
                             ], 0)

            # detect the sky area in each differential combination

            for i in range(len(differential_science)):

                # calculate the distribution of the flux and fit a gaussian, do not consider pixels at the edges

                frame = differential_science[i][10:-10, 10:-10]
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    frame_mean, frame_std = plc.fit_one_d_distribution_gaussian(frame, step=10.0)[2][-2:]
                frame_std = np.abs(frame_std)

                check_array = (np.abs(frame - frame_mean) < sky_detection_limit.value * frame_std)

                if np.sum(check_array) == 0:
                    frame_mean = np.median(frame)
                    frame_std = np.median(np.abs(frame - frame_mean))
                    check_array = (np.abs(frame - frame_mean) < sky_detection_limit.value * frame_std)

                differential_science[i] = check_array

            # detect the final sky area from all the differential combination

            sky_area = np.where(np.all(np.array(differential_science), 0))
            if len(sky_area[0]) == 0:
                sky_area = np.where(np.sum(np.array(differential_science), 0) > 0.75 * len(np.array(differential_science)))
            sky_area = (sky_area[0] + 10, sky_area[1] + 10)

            # save the sky area in the fits file (0 is where the pixels is illuminated only by the sky background)

            sky_map_array = np.ones_like(fits[1].data)
            sky_map_array[sky_area] = 0
            sky_frame.set(sky_map_array)
            sky_frame.to_fits(fits)

            # save the sky detection limit in the fits file

            sky_detection_limit.to_fits(fits)

        else:

            sky_frame.from_fits(fits)
            sky_area = np.where(sky_frame.value == 0)

        # iterate over the NDRs

        for i, j in plc.fits_sci_err(fits):

            # get the NDR and its error-array from the data file

            science = np.array(fits[i].data, dtype=float)
            error = np.array(fits[j].data, dtype=float)

            # calculate the sky background level

            sky_background_level.set(np.median((science / master_sky)[sky_area]))

            # subtract the master-sky scaled to the sky background level

            science -= master_sky * sky_background_level.value

            # propagate the uncertainties in the error-array

            # TODO: Consider using the standard deviation of the sky ratio instead of the sky photon noise

            error = np.sqrt(error * error + np.abs(master_sky * sky_background_level.value))

            # save the sky background level in the fits file

            sky_background_level.to_fits(fits, position=i)

            # update the NDR and its error-array in the data file

            fits[i].data = science
            fits[j].data = error

        # update counter

        counter.update()

    # return the updated version of the input data

    return input_data


def flat(input_data):
    """
    Flat-field correction.

    1) Corrects each NDR for the flat-field.

    by dividing it by the wavelength-dependent flat-field calculated ... .

    2) Propagates the uncertainties in the error-arrays.

    Parameters
    ----------
    input_data : HDUList or DataSet object
        input data, either a single image or a complete data set

    Returns
    -------
    input_data : HDUList or DataSet object
        updated version of the input_data object, corrected for gain variations

    """

    # load pipeline and calibration variables to be used

    normalised_wavelength_frame = variables.normalised_wavelength_frame

    flat_field_coefficient_1 = calibrations.flat_field_coefficient_1.match(input_data)
    flat_field_coefficient_2 = calibrations.flat_field_coefficient_2.match(input_data)
    flat_field_coefficient_3 = calibrations.flat_field_coefficient_3.match(input_data)
    flat_field_coefficient_4 = calibrations.flat_field_coefficient_4.match(input_data)

    # initiate counter

    counter = PipelineCounter('Flat', len(input_data.spectroscopic_images))

    # iterate over the list of HDUList objects included in the input data

    for fits in input_data.spectroscopic_images:

        # get the wavelength of each pixel from the fits file

        normalised_wavelength_frame.from_fits(fits)

        # calculate the wavelength-dependent flat-field

        flat_field = (flat_field_coefficient_1 +
                      flat_field_coefficient_2 * normalised_wavelength_frame.value +
                      flat_field_coefficient_3 * (normalised_wavelength_frame.value ** 2) +
                      flat_field_coefficient_4 * (normalised_wavelength_frame.value ** 3))

        # exclude outliers

        flat_field = np.where(flat_field < 0.01, 1, flat_field)
        flat_field = np.where(flat_field > 1.1, 1, flat_field)

        # iterate over the NDRs

        for i, j in plc.fits_sci_err(fits):

            # get the NDR and its error-array from the data file

            science = np.array(fits[i].data, dtype=float)
            error = np.array(fits[j].data, dtype=float)

            # divide by the flat field

            science /= flat_field

            # propagate the uncertainties in the error-array

            error /= flat_field

            # update the NDR and its error-array in the data file

            fits[i].data = science
            fits[j].data = error

        # update counter

        counter.update()

    # return the updated version of the input data

    return input_data


def bpcr(input_data, cr_detection_limit=None, cr_neighbours=None, use_bpcr_fast_mode=None, splitting=False):
    """
    Bad-pixels and cosmic-rays correction.

    Corrects each read for the bad pixels and the cosmic rays.
    The bad pixels are given by the pixels calibration file while the cosmic rays are detected.

    For each pixel, two flags are calculated;
    the median absolute difference from the N (cr_neighbours) horizontally neighbouring pixels (x-flag) and
    the median absolute difference from the N (cr_neighbours) vertically neighbouring pixels (y-flag).
    Given the cosmic rays detection threshold c (cr_detection_limit),
    if a pixel's x-flag is c times larger than the median x-flag in the column (N closest pixels in the column) and
    it's y-flag is c times larger than the median y-flag in the row (N closest pixels in the row),
    it is identified as frame cosmic ray.

    Parameters
    ----------
    input_data : HDUList or DataSet object
        input data, either frame single image or frame complete data set

    cr_detection_limit : float
        detection threshold for cosmic rays

    cr_neighbours : integer
        number of neighbouring pixels from which the x and y flags are calculated

    Returns
    -------
    input_data : HDUList or DataSet object
        updated version of the input_data object, corrected for bad pixels and cosmic rays

    """

    # load pipeline and calibration variables to be used

    cr_neighbours = variables.cr_neighbours.custom(cr_neighbours)
    cr_detection_limit = variables.cr_detection_limit.custom(cr_detection_limit)
    use_bpcr_fast_mode = variables.use_bpcr_fast_mode.custom(use_bpcr_fast_mode)

    bpcr_map = variables.bpcr_map.custom()

    bad_pixels = calibrations.bad_pixels.match(input_data)

    # initiate counter

    counter = PipelineCounter('BPs & CRs', len(input_data.spectroscopic_images))

    # iterate over the list of HDUList objects included in the input data

    for fits in input_data.spectroscopic_images:

        if not splitting:

            frame = np.array(fits[1].data, dtype=float)

            # calculate the x and y flags from the median absolute difference between each pixel
            # and each N neighbours (N = cr_neighbours)

            x_flag = []
            y_flag = []

            for i in range(1, int(1 + cr_neighbours.value / 2)):
                x_flag.append(np.abs(frame - np.roll(frame, i, 1)))
                x_flag.append(np.abs(frame - np.roll(frame, -i, 1)))
                y_flag.append(np.abs(frame - np.roll(frame, i, 0)))
                y_flag.append(np.abs(frame - np.roll(frame, -i, 0)))

            x_flag = np.median(x_flag, 0)
            y_flag = np.median(y_flag, 0)

            # calculate the detection limit as me median x_flag per column (from the N closest pixels in the column)
            # and the median y-flag per row (from the N closest pixels in the row)

            x_detection_limit = []
            y_detection_limit = []

            for i in range(1, int(1 + cr_neighbours.value / 2)):
                x_detection_limit.append(np.roll(x_flag, i, 1))
                x_detection_limit.append(np.roll(x_flag, -i, 1))
                y_detection_limit.append(np.roll(y_flag, i, 0))
                y_detection_limit.append(np.roll(y_flag, -i, 0))

            x_detection_limit = cr_detection_limit.value * np.median(x_detection_limit, 0)
            y_detection_limit = cr_detection_limit.value * np.median(y_detection_limit, 0)

            # characterise as cosmic rays only the pixes which have both flags above the respective detection limit

            cr_test = np.where((x_flag > x_detection_limit) & (y_flag > y_detection_limit), 1, 0)
            cr_test = np.where(cr_test == 1)

            # in the the bad pixels table the bad pixels are characterised by the number -1
            # characterise the cosmic rays by the number 1
            # and save the final bad pixels and cosmic rays map it in the fts file

            bpcr_map_array = np.ones_like(bad_pixels) * bad_pixels
            bpcr_map_array[cr_test] = 1
            bpcr_map.set(bpcr_map_array)
            bpcr_map.to_fits(fits)

            # save process information in the fts file

            cr_neighbours.to_fits(fits)
            cr_detection_limit.to_fits(fits)

        else:

            bpcr_map.from_fits(fits)

        # scan or no scan

        data = np.sum(fits[1].data, 0)
        model = plc.box(np.arange(len(data)), len(data) / 2, np.max(data), 45., 40.)
        dx = np.argmax(np.convolve(data, model)) - np.argmax(np.convolve(model, model))
        x_lim1, x_lim2 = int(len(data) / 2 + dx - 55), int(len(data) / 2 + dx + 55)

        # detect the approximate vertical position of the final spectrum

        data = fits[1].data
        y1 = 5 + np.median(np.argmax(data[5:, x_lim1:x_lim2] - data[:-5, x_lim1:x_lim2], 0))
        y2 = np.median(np.argmin(data[5:, x_lim1:x_lim2] - data[:-5, x_lim1:x_lim2], 0))
        if abs(y2 - y1) <= 1:
            y1 = float(2 + np.median(np.argmax(data[2:, x_lim1:x_lim2] - data[:-2, x_lim1:x_lim2], 0)))
            y2 = float(np.median(np.argmin(data[2:, x_lim1:x_lim2] - data[:-2, x_lim1:x_lim2], 0)))
        final_y_lim1, final_y_lim2 = np.sort([int(round(y1)), int(round(y2))])

        if abs(final_y_lim1 - final_y_lim2) > 5:
            final_scan = True
        else:
            final_scan = False

        # correct each read for the bad pixels and the cosmic rays

        if len(plc.fits_sci_err(fits)) == 1:
            corr = [plc.fits_sci_err(fits)[0]]
        else:
            if use_bpcr_fast_mode.value:
                corr = [plc.fits_sci_err(fits)[0], plc.fits_sci_err(fits)[-1]]
            else:
                corr = plc.fits_sci_err(fits)

        for i, j in corr:

            # get the NDR and its error-array from the data file

            science = np.array(fits[i].data, dtype=float)
            science_old = np.array(fits[i].data, dtype=float)
            error = np.array(fits[j].data, dtype=float)

            if final_scan:

                # find the clean pixels location and values

                grid_x, grid_y = np.meshgrid(np.arange(len(science)), np.arange(len(science[0])))
                clean_pixs = np.where(bpcr_map.value == 0)
                points = np.swapaxes(np.roll(clean_pixs, 1, 0), 0, 1)
                values = fits[i].data[clean_pixs].flatten()

                # replace the frame with the interpolated values
                # after this process the clean pixels remain the same
                # and the bad pixels and cosmic rays are replaced

                # TODO: this function is slow, try and write a faster process using np.arrays

                science = griddata(points, values, (grid_x, grid_y), method='cubic')

            else:
                bpcr_map.value[:, 0] = 0
                bpcr_map.value[:, -1] = 0
                for fits_line in range(len(fits[i].data)):
                    clean_pixs = np.where(bpcr_map.value[fits_line] == 0)
                    points = clean_pixs[0]
                    values = fits[i].data[fits_line][clean_pixs]
                    line_model = interp1d(points, values)
                    science[fits_line] = line_model(np.arange(len(science[fits_line])))

            # propagate the uncertainties in the error array assuming that the difference between
            # the original the final values is and additive correction

            # TODO: This calculation is not the most optimal. A better approach would be to assume that the
            # TODO: interpolated value is a new measurement with an uncertainty originated from the interpolation.
            # TODO: The photon noise (square root of the interpolated value) should be added to that uncertainty to
            # TODO: calculate the final uncertainty.

            error = np.sqrt(error ** 2 + np.abs(science - science_old))

            # update the NDR and its error-array in the data file

            fits[i].data = science
            fits[j].data = error

        # update counter

        counter.update()

    return input_data



from setuptools import setup
import codecs
import os
import glob

name = 'iraclis'
version = open(os.path.join(os.path.abspath(os.path.dirname(__file__)), name, '__version__.txt')).read()
description = 'Analysis pipeline for HST/WFC3 spectroscopic observations of exoplanet transits and eclipses'
url = 'https://github.com/ucl-exoplanets/Iraclis'
install_requires = ['docopt', 'pylightcurve>=3.0.5', 'numpy>=1.19.2', 'matplotlib>=3.3.2', 'scipy>=1.5.2',
                    'emcee>=3.0.2', 'sklearn', 'astropy>=4.2']
entry_point = '__main__:console'

os.chdir(os.path.abspath(os.path.dirname(__file__)))

subdirs_to_include = []
for x in os.walk(name):
    if os.path.isdir(x[0]):
        if x[0] != name:
            subdirs_to_include.append(x[0])

files_to_include = []
for x in glob.glob(os.path.join(name, '*')):
    if os.path.isfile(x):
        if x.split('.')[-1] not in ['py']:
            files_to_include.append(os.path.join(name, os.path.split(x)[1]))

files_to_include.append('README.md')
files_to_include.append('LICENSE')
files_to_include.append('readme.md')
files_to_include.append('licence')

w = open('MANIFEST.in', 'w')
for i in subdirs_to_include:
    w.write('include ' + os.path.join(i, '*') + ' \n')

for i in files_to_include:
    w.write('include ' + i + ' \n')

w.close()

try:
    with codecs.open('README.md', encoding='utf-8') as f:
        long_description = f.read()
except:
    with codecs.open('readme.md', encoding='utf-8') as f:
        long_description = f.read()

setup(
    name=name,
    version=version,
    description=description,
    long_description=long_description,
    url=url,
    author='Angelos Tsiaras',
    author_email='aggelostsiaras@gmail.com',
    license='Creative Commons Attribution 4.0 International License',
    classifiers=['Development Status :: 4 - Beta',
                 'Environment :: Console',
                 'Intended Audience :: Science/Research',
                 'Topic :: Scientific/Engineering :: Astronomy',
                 'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
                 'Operating System :: MacOS :: MacOS X',
                 'Programming Language :: Python :: 3.7',
                 ],
    entry_points={'console_scripts': ['{0} = {0}.{1}'.format(name, entry_point)]},
    packages=[name],
    install_requires=install_requires,
    include_package_data=True,
    zip_safe=False,
)
